/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "client/src/cfg.hh"

#include <syslog.h>

#include <cstdlib>
#include <fstream>
#include <iostream>
#include <regex>
#include <unordered_map>

namespace Fred {
namespace Zone {
namespace Client {
namespace Cfg {

namespace {

decltype(auto) make_severity(const std::string& str)
{
    static const std::unordered_map<std::string, LibLog::Level> str2level =
    {
        { "critical", LibLog::Level::critical },
        { "error", LibLog::Level::error },
        { "warning", LibLog::Level::warning },
        { "info", LibLog::Level::info },
        { "debug", LibLog::Level::debug },
        { "trace", LibLog::Level::trace }
    };
    const auto level_itr = str2level.find(str);
    if (level_itr != str2level.end())
    {
        return level_itr->second;
    }
    throw Exception("\"" + str + "\" not a valid severity");
}

constexpr int fake_argc = -1;
constexpr const char* const fake_argv[] = { nullptr };

}//namespace Fred::Zone::Client::Cfg::{anonymous}

const Options& Options::get()
{
    return init(fake_argc, fake_argv);
}

const Options& Options::init(int argc, const char* const* argv)
{
    static const Options* singleton_ptr = nullptr;
    const bool init_requested = (argc != fake_argc) || (argv != fake_argv);
    const bool first_run = singleton_ptr == nullptr;
    if (first_run)
    {
        if (!init_requested)
        {
            throw std::runtime_error("First call of Cfg::Options::init must contain valid arguments");
        }
        static const Options singleton{argc, argv};
        singleton_ptr = &singleton;
    }
    else if (init_requested)
    {
        throw std::runtime_error("Only first call of Cfg::Options::init can contain valid arguments");
    }
    return *singleton_ptr;
}

namespace {

constexpr struct OptionNames
{
    static constexpr struct ZoneService
    {
        static constexpr auto connect_to = "zone_service.connect_to";
    } zone_service = {};
    static constexpr struct Generate
    {
        static constexpr auto zones = "generate.zones";
    } generate = {};
} option{};

static constexpr unsigned terminal_width = 120;

decltype(auto) make_generic_options(Options::Generic& generic)
{
    boost::program_options::options_description generic_options{"Generic options", terminal_width};
    const auto set_backup_directory = [&](const std::string& value) { generic.zone_common.backup_directory = value; };
    const auto set_bind_config_file_name = [&](const std::string& value) { generic.zone_common.bind_config_file_name = value; };
    const auto set_config_file_name = [&](const std::string& value) { generic.config_file_name = value; };
    const auto set_footer_file_name = [&](const std::string& value) { generic.zone_common.footer_file_name = value; };
    const auto set_header_file_name = [&](const std::string& value) { generic.zone_common.header_file_name = value; };
    const auto set_max_changes = [&](const std::string& value)
    {
        if (!value.empty() && (value.back() == '%'))
        {
            generic.zone_common.max_changes_relative = std::stod(value.substr(0, value.length() - 1)) / 100;
        }
        else
        {
            generic.zone_common.max_changes_absolute = std::stoul(value);
        }
    };
    const auto set_post_hook_command = [&](const std::string& value) { generic.zone_common.post_hook_command = value; };
    const auto set_zone_directory = [&](const std::string& value) { generic.zone_common.zone_directory = value; };
    generic_options.add_options()
        ("help,h", "produce help message")
        ("backupdir,b", boost::program_options::value<std::string>()->notifier(set_backup_directory),
         "directory used for zone file backups (default is the zonedir directory)")
        ("bind-conf,g", boost::program_options::value<std::string>()->notifier(set_bind_config_file_name),
         "path to file where bind configuration will be generated")
        ("config,c",
         boost::program_options::value<std::string>()->default_value(std::string{"client-zone.conf"})->notifier(set_config_file_name),
         "configuration file of this client")
        ("footer,t", boost::program_options::value<std::string>()->notifier(set_footer_file_name),
         "path to file which will be appended to zone file")
        ("header,e", boost::program_options::value<std::string>()->notifier(set_header_file_name),
         "path to file which will be prepended to zone file")
        ("maxchanges,m", boost::program_options::value<std::string>()->notifier(set_max_changes),
         "maximal number of changed lines compared to previously generated zone; if you append '%' "
         "to the number, then it means a procentual change")
        ("nobackups,o", boost::program_options::value<bool>(&generic.zone_common.no_backups)->zero_tokens(),
         "disable creation of zone backups")
        ("post-hook-command,p", boost::program_options::value<std::string>()->notifier(set_post_hook_command),
         "execute specified command after each zone is generated (see README)")
        ("post-hook-argument,a", boost::program_options::value<std::vector<std::string>>(&generic.zone_common.post_hook_arguments),
         "arguments for the post-hook-command")
        ("zonedir,z", boost::program_options::value<std::string>()->notifier(set_zone_directory),
         "directory where the zone file(s) should be generated (default is current directory)");
    return generic_options;
}

struct CommonLogOptions
{
    enum class Device
    {
        console,
        file,
        syslog
    };
    static decltype(auto) make_device(const std::string& device)
    {
        static const std::unordered_map<std::string, Device> str2device =
            {
                { "console", Device::console },
                { "file", Device::file },
                { "syslog", Device::syslog }
            };
        const auto device_itr = str2device.find(device);
        if (device_itr != str2device.end())
        {
            return device_itr->second;
        }
        throw Exception{"\"" + device + "\" is not a valid device"};
    }
    std::set<Device> devices;
    boost::optional<LibLog::Level> min_severity;
    boost::optional<bool> log_grpc_library_events;
    void set_devices(const std::vector<std::string>& str_devices)
    {
        std::for_each(
            str_devices.begin(),
            str_devices.end(),
            [&](const std::string& str_device)
            {
                devices.insert(make_device(str_device));
            });
    }
    void set_min_severity(const std::string& severity)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(severity);
    }
    void set_log_grpc_library_events(bool value)
    {
        log_grpc_library_events = value;
    }
};

decltype(auto) make_log_options(CommonLogOptions& common_log_options)
{
    boost::program_options::options_description options_description{"Logging options", terminal_width};
    const auto set_devices = [&](const std::vector<std::string>& devices) { common_log_options.set_devices(devices); };
    const auto set_min_severity = [&](const std::string& severity) { common_log_options.set_min_severity(severity); };
    const auto set_log_grpc_library_events = [&](bool to_log) { common_log_options.set_log_grpc_library_events(to_log); };
    options_description.add_options()
        ("log.device", boost::program_options::value<std::vector<std::string>>()->multitoken()->notifier(set_devices), "where to log (console/file/syslog)")
        ("log.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace")
        ("log.grpc_library",
         boost::program_options::value<bool>()->zero_tokens()->notifier(set_log_grpc_library_events),
         "log gRPC library events");
    return options_description;
}

decltype(auto) make_output_stream(const std::string& str)
{
    static const std::unordered_map<std::string, LibLog::Sink::ConsoleSinkConfig::OutputStream> str2output_stream =
        {
            { "stderr", LibLog::Sink::ConsoleSinkConfig::OutputStream::stderr },
            { "stdout", LibLog::Sink::ConsoleSinkConfig::OutputStream::stdout }
        };
    const auto output_stream_itr = str2output_stream.find(str);
    if (output_stream_itr != str2output_stream.end())
    {
        return output_stream_itr->second;
    }
    throw Exception("\"" + str + "\" not a valid output stream");
}

decltype(auto) make_color_mode(const std::string& str)
{
    static const std::unordered_map<std::string, LibLog::ColorMode> str2color_mode =
        {
            { "always", LibLog::ColorMode::always },
            { "auto", LibLog::ColorMode::automatic },
            { "never", LibLog::ColorMode::never }
        };
    const auto color_mode_itr = str2color_mode.find(str);
    if (color_mode_itr != str2color_mode.end())
    {
        return color_mode_itr->second;
    }
    throw Exception("\"" + str + "\" not a valid color mode");
}

struct ConsoleLogOptions
{
    boost::optional<LibLog::Level> min_severity;
    boost::optional<LibLog::Sink::ConsoleSinkConfig::OutputStream> output_stream;
    boost::optional<LibLog::ColorMode> color_mode;
    void set_min_severity(const std::string& value)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(value);
    }
    void set_output_stream(const std::string& value)
    {
        if (output_stream != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        output_stream = make_output_stream(value);
    }
    void set_color_mode(const std::string& value)
    {
        if (color_mode != boost::none)
        {
            throw Exception{"color mode can be defined at most once"};
        }
        color_mode = make_color_mode(value);
    }
};

decltype(auto) make_log_console_options(ConsoleLogOptions& console_log_options)
{
    boost::program_options::options_description options_description{"Logging on console options", terminal_width};
    const auto set_min_severity = [&](const std::string& severity) { console_log_options.set_min_severity(severity); };
    const auto set_output_stream = [&](const std::string& output_stream) { console_log_options.set_output_stream(output_stream); };
    const auto set_color_mode = [&](const std::string& color_mode) { console_log_options.set_color_mode(color_mode); };
    options_description.add_options()
        ("log.console.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace")
        ("log.console.output_stream",
         boost::program_options::value<std::string>()->notifier(set_output_stream),
         "where to log (stderr/stdout), default is stderr")
        ("log.console.color_mode",
         boost::program_options::value<std::string>()->notifier(set_color_mode),
         "how to colorize output (always/auto/never), default is never");
    return options_description;
}

struct FileLogOptions
{
    boost::optional<std::string> file_name;
    boost::optional<LibLog::Level> min_severity;
    void set_file_name(const std::string& name)
    {
        if (file_name != boost::none)
        {
            throw Exception{"file name can be defined at most once"};
        }
        file_name = name;
    }
    void set_min_severity(const std::string& severity)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(severity);
    }
};

decltype(auto) make_log_file_options(FileLogOptions& file_log_options)
{
    boost::program_options::options_description options_description{"Logging into file options", terminal_width};
    const auto set_file_name = [&](const std::string& file_name) { file_log_options.set_file_name(file_name); };
    const auto set_min_severity = [&](const std::string& severity) { file_log_options.set_min_severity(severity); };
    options_description.add_options()
        ("log.file.file_name", boost::program_options::value<std::string>()->notifier(set_file_name), "what file to log into")
        ("log.file.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace");
    return options_description;
}

struct SyslogLogOptions
{
    boost::optional<std::string> ident;
    boost::optional<int> facility_local_offset;
    boost::optional<LibLog::Level> min_severity;
    void set_ident(const std::string& value)
    {
        if (ident != boost::none)
        {
            throw Exception{"ident can be defined at most once"};
        }
        ident = value;
    }
    void set_facility(int value)
    {
        if ((value < 0) || (7 < value))
        {
            throw Exception("facility out of range [0, 7]");
        }
        if (facility_local_offset != boost::none)
        {
            throw Exception{"facility can be defined at most once"};
        }
        facility_local_offset = value;
    }
    void set_min_severity(const std::string& severity)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(severity);
    }
};

decltype(auto) make_log_syslog_options(SyslogLogOptions& syslog_log_options)
{
    boost::program_options::options_description options_description{"Logging into syslog options", terminal_width};
    const auto set_ident = [&](const std::string& ident) { syslog_log_options.set_ident(ident); };
    const auto set_facility = [&](int facility) { syslog_log_options.set_facility(facility); };
    const auto set_min_severity = [&](const std::string& severity) { syslog_log_options.set_min_severity(severity); };
    options_description.add_options()
        ("log.syslog.ident", boost::program_options::value<std::string>()->notifier(set_ident), "what ident to log with (default is empty string)")
        ("log.syslog.facility",
         boost::program_options::value<int>()->notifier(set_facility),
         "what LOG_LOCALx facility to log with (x in range 0..7, default means facility LOG_USER)")
        ("log.syslog.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace");
    return options_description;
}

decltype(auto) make_zone_service_options(Options::ZoneService& service)
{
    boost::program_options::options_description service_options{"Zone service options", terminal_width};
    service_options.add_options()
        (option.zone_service.connect_to, boost::program_options::value<std::string>(&service.connect_to), "server's listen address");
    return service_options;
}

decltype(auto) make_generate_options(Options::Generate& generate)
{
    boost::program_options::options_description service_options{"Generated zones options", terminal_width};
    service_options.add_options()
        (option.generate.zones, boost::program_options::value<std::vector<std::string>>(&generate.zones),
         "list of zone names to be generated; if no zones are specified at all, list of zones is downloaded from server");
    return service_options;
}

decltype(auto) make_zone_options(Options::ParticularZones::value_type& particular_zone, Options::NamesStorage& storage)
{
    boost::program_options::options_description zone_options{"Particular " + particular_zone.first + " zone options", terminal_width};
    auto& zone = particular_zone.second;
    const auto set_backup_directory = [&](const std::string& directory) { zone.backup_directory = directory; };
    const auto set_zone_directory = [&](const std::string& directory) { zone.zone_directory = directory; };
    const auto set_max_changes = [&](const std::string& max_changes)
    {
        if (!max_changes.empty() && (max_changes.back() == '%'))
        {
            zone.max_changes_relative = std::stod(max_changes.substr(0, max_changes.length() - 1)) / 100;
        }
        else
        {
            zone.max_changes_absolute = std::stoul(max_changes);
        }
    };
    const auto set_bind_config_file_name = [&](const std::string& value) { zone.bind_config_file_name = value; };
    const auto set_footer_file_name = [&](const std::string& value) { zone.footer_file_name = value; };
    const auto set_header_file_name = [&](const std::string& value) { zone.header_file_name = value; };
    const auto set_post_hook_command = [&](const std::string& value) { zone.post_hook_command = value; };
    const auto add_zone_option = [&](const char* option) { storage.push_back("zone:\"" + particular_zone.first + "\"." + option); };
    add_zone_option("backupdir");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_backup_directory),
         "directory path of the backup of the old zone file (default: same as zonedir)");
    add_zone_option("zonedir");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_zone_directory),
         "directory where new zone files should be generated (default is current directory)");
    add_zone_option("maxchanges");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_max_changes),
         "maximum of changed records between the old and the new zone file, if terminated by the '%' character it does not mean "
         "absolute change but relative change; this option applies only if the previous zone file is found.");
    add_zone_option("bind-conf");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_bind_config_file_name),
        "path to file where bind configuration will be generated");
    add_zone_option("footer");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_footer_file_name),
        "path to file which will be appended to zone file");
    add_zone_option("header");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_header_file_name),
        "path to file which will be prepended to zone file");
    add_zone_option("nobackups");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<bool>(&zone.no_backups)->zero_tokens(),
        "disable creation of zone backups");
    add_zone_option("post-hook-command");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::string>()->notifier(set_post_hook_command),
        "execute specified command after each zone is generated (see README)");
    add_zone_option("post-hook-argument");
    zone_options.add_options()
        (storage.back().c_str(), boost::program_options::value<std::vector<std::string>>(&zone.post_hook_arguments),
        "arguments for the post-hook-command");
    return zone_options;
}

void collect_log_options(
        const CommonLogOptions& common_log_options,
        const ConsoleLogOptions& console_log_options,
        const FileLogOptions& file_log_options,
        const SyslogLogOptions& syslog_log_options,
        Options::Log& log_options)
{
    log_options.log_grpc_library_events = (common_log_options.log_grpc_library_events != boost::none) && *common_log_options.log_grpc_library_events;

    static constexpr auto default_log_min_severity = LibLog::Level::critical;
    const auto common_log_min_severity = common_log_options.min_severity != boost::none ? *common_log_options.min_severity
                                                                                        : default_log_min_severity;
    std::for_each(
        common_log_options.devices.begin(),
        common_log_options.devices.end(),
        [&](CommonLogOptions::Device device)
        {
            switch (device)
            {
                case CommonLogOptions::Device::console:
                {
                    Options::Log::Console console;
                    console.min_severity = console_log_options.min_severity != boost::none ? *console_log_options.min_severity
                                                                                           : common_log_min_severity;
                    static constexpr auto default_output_stream = LibLog::Sink::ConsoleSinkConfig::OutputStream::stderr;
                    console.output_stream = console_log_options.output_stream != boost::none ? *console_log_options.output_stream
                                                                                             : default_output_stream;
                    static constexpr auto default_color_mode = LibLog::ColorMode::never;
                    console.color_mode = console_log_options.color_mode != boost::none ? *console_log_options.color_mode
                                                                                       : default_color_mode;
                    log_options.console = console;
                    return;
                }
                case CommonLogOptions::Device::file:
                {
                    if (file_log_options.file_name == boost::none)
                    {
                        throw MissingOption{"missing option: 'log.file.file_name'"};
                    }
                    Options::Log::File log_file;
                    log_file.file_name = *file_log_options.file_name;
                    log_file.min_severity = file_log_options.min_severity != boost::none ? *file_log_options.min_severity
                                                                                         : common_log_min_severity;
                    log_options.file = log_file;
                    return;
                }
                case CommonLogOptions::Device::syslog:
                {
                    Options::Log::Syslog syslog;
                    static const std::string default_ident = "";
                    syslog.ident = syslog_log_options.ident != boost::none ? *syslog_log_options.ident
                                                                           : default_ident;
                    static constexpr int default_facility = LOG_USER;
                    static const auto make_facility = [](int offset)
                    {
                        switch (offset)
                        {
                            case 0: return LOG_LOCAL0;
                            case 1: return LOG_LOCAL1;
                            case 2: return LOG_LOCAL2;
                            case 3: return LOG_LOCAL3;
                            case 4: return LOG_LOCAL4;
                            case 5: return LOG_LOCAL5;
                            case 6: return LOG_LOCAL6;
                            case 7: return LOG_LOCAL7;
                        }
                        return default_facility;
                    };
                    syslog.facility = syslog_log_options.facility_local_offset != boost::none ? make_facility(*syslog_log_options.facility_local_offset)
                                                                                              : default_facility;
                    syslog.min_severity = syslog_log_options.min_severity != boost::none ? *syslog_log_options.min_severity
                                                                                         : common_log_min_severity;
                    static constexpr int default_options = LOG_CONS | LOG_ODELAY;
                    syslog.options = default_options;
                    log_options.syslog = syslog;
                    return;
                }
            }
            throw Exception("Invalid device");
        });
}

}//namespace Fred::Zone::Client::Cfg::{anonymous}

Options::Options(int argc, const char* const* argv)
{
    CommonLogOptions common_log_options;
    ConsoleLogOptions console_log_options;
    FileLogOptions file_log_options;
    SyslogLogOptions syslog_log_options;
    const boost::program_options::options_description generic_options = make_generic_options(generic);
    const boost::program_options::options_description log_options = make_log_options(common_log_options);
    const boost::program_options::options_description log_console_options = make_log_console_options(console_log_options);
    const boost::program_options::options_description log_file_options = make_log_file_options(file_log_options);
    const boost::program_options::options_description log_syslog_options = make_log_syslog_options(syslog_log_options);
    const boost::program_options::options_description service_options = make_zone_service_options(zone_service);
    const boost::program_options::options_description generate_options = make_generate_options(generate);

    boost::program_options::options_description command_line_options{"fred-zone-generator options", terminal_width};
    command_line_options
        .add(generic_options)
        .add(log_options)
        .add(log_console_options)
        .add(log_file_options)
        .add(log_syslog_options)
        .add(service_options)
        .add(generate_options);
    boost::program_options::options_description config_file_options;
    config_file_options
        .add(log_options)
        .add(log_console_options)
        .add(log_file_options)
        .add(log_syslog_options)
        .add(service_options)
        .add(generate_options);

    boost::program_options::variables_map variables_map{};
    boost::program_options::positional_options_description pos_options{};
    pos_options.add(option.generate.zones, 1);
    const auto has_option = [&](const char* option) { return 0 < variables_map.count(option); };
    try
    {
        boost::program_options::store(
                boost::program_options::command_line_parser(argc, argv)
                    .options(command_line_options)
                    .positional(pos_options)
                    .run(),
                variables_map);
    }
    catch (const boost::program_options::unknown_option& unknown_option)
    {
        std::ostringstream out;
        out << unknown_option.what() << "\n\n" << command_line_options;
        throw UnknownOption(out.str());
    }
    boost::program_options::notify(variables_map);

    if (has_option("help"))
    {
        std::ostringstream out;
        out << command_line_options;
        throw HelpRequested(out.str());
    }
    const bool config_file_name_presents = generic.config_file_name != boost::none;
    if (config_file_name_presents)
    {
        const auto get_config_file = [&]()
        {
            std::ifstream config_file{*generic.config_file_name};
            if (!config_file)
            {
                throw Exception("can not open config file \"" + *generic.config_file_name + "\"");
            }
            return config_file;
        };
        try
        {
            constexpr auto allow_unregistered = true;
            auto config_file = get_config_file();
            auto parsed_config_file = boost::program_options::parse_config_file(
                    config_file,
                    config_file_options,
                    allow_unregistered);
            std::list<boost::program_options::options_description> zones_options;
            std::for_each(begin(parsed_config_file.options), end(parsed_config_file.options), [&](const auto& option)
            {
                if (option.unregistered)
                {
                    constexpr auto regex_options = std::regex_constants::extended | std::regex_constants::optimize;
                    //zone:"cz".option
                    const auto regex = std::regex("^zone:\"([^\"]+)\"\\.", regex_options);
                    std::smatch match_iter{};
                    if (!std::regex_search(option.string_key, match_iter, regex))
                    {
                        throw UnknownOption{"unknown option " + option.string_key};
                    }
                    const auto zone_fqdn = match_iter[1].str();
                    const bool zone_first_occurrence = zones.find(zone_fqdn) == zones.end();
                    if (zone_first_occurrence)
                    {
                        zones_options.push_back(make_zone_options(*zones.insert(std::make_pair(zone_fqdn, Zone{})).first, names_storage_));
                        config_file_options.add(zones_options.back());
                    }
                }
            });
            constexpr auto disable_unregistered = !allow_unregistered;
            config_file = get_config_file();
            parsed_config_file = boost::program_options::parse_config_file(
                    config_file,
                    config_file_options,
                    disable_unregistered);
            boost::program_options::store(parsed_config_file, variables_map);
        }
        catch (const boost::program_options::unknown_option& unknown_option)
        {
            std::ostringstream out;
            out << unknown_option.what() << "\n\n" << command_line_options;
            throw UnknownOption(out.str());
        }
        boost::program_options::notify(variables_map);
    }

    collect_log_options(
            common_log_options,
            console_log_options,
            file_log_options,
            syslog_log_options,
            log);

    const auto required = [&](const char* variable)
    {
        if (variables_map.count(variable) == 0)
        {
            throw MissingOption{"missing option: '" + std::string{variable} + "'"};
        }
    };

    required(option.zone_service.connect_to);
}

HelpRequested::HelpRequested(std::string msg)
    : msg_{std::move(msg)}
{ }

const char* HelpRequested::what() const noexcept
{
    return msg_.c_str();
}

Exception::Exception(std::string msg)
    : std::runtime_error{std::move(msg)}
{ }

UnknownOption::UnknownOption(std::string msg)
    : Exception{std::move(msg)}
{ }

MissingOption::MissingOption(std::string msg)
    : Exception{std::move(msg)}
{ }

void handle_cli_args(int argc, char** argv)
{
    try
    {
        Cfg::Options::init(argc, argv);
    }
    catch (const Cfg::HelpRequested& help)
    {
        std::cout << help.what() << std::endl;
        ::exit(EXIT_SUCCESS);
    }
    catch (const Cfg::UnknownOption& unknown_option)
    {
        std::cerr << unknown_option.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (const Cfg::MissingOption& missing_option)
    {
        std::cerr << missing_option.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (const Cfg::Exception& e)
    {
        std::cerr << "Configuration problem: " << e.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        ::exit(EXIT_FAILURE);
    }
}

}//namespace Fred::Zone::Client::Cfg
}//namespace Fred::Zone::Client
}//namespace Fred::Zone
}//namespace Fred
