/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "client/src/generate_zone.hh"
#include "client/src/unwrap_get_zone_data.hh"

#include "liblog/liblog.hh"

#include <algorithm>
#include <fstream>
#include <functional>
#include <iostream>
#include <memory>
#include <regex>
#include <stdexcept>
#include <string>
#include <tuple>
#include <utility>

namespace Fred {
namespace Zone {
namespace Client {

namespace {

class GetZoneDataVisitorImpl : public GetZoneDataVisitor
{
public:
    explicit GetZoneDataVisitorImpl(std::ostream& out, ZoneData* zone_data);
    std::uint32_t get_soa_id() const noexcept;
private:
    void operator()(Common common) override;
    void operator()(Record record) override;
    void operator()(NotFound exception) override;
    void operator()(Exception exception) override;

    std::ostream& data_section(const std::string& owner, const std::string& record_type);
    std::ostream& zone_section(const std::string& owner, const std::string& record_type);

    std::ostream& out_;
    ZoneData* const zone_data_;
    std::string origin_;
    std::uint32_t soa_id_;
    bool insert_horizontal_space_;
};

}//namespace Fred::Zone::Client::{anonymous}

std::uint32_t generate_zone(const std::string& zone, const Cfg::Options::ZoneService& zone_service_options, std::ostream& out, ZoneData* zone_data)
{
    auto visitor = GetZoneDataVisitorImpl{out, zone_data};
    unwrap_get_zone_data(zone_service_options, zone, visitor);
    return visitor.get_soa_id();
}

namespace {

GetZoneDataVisitorImpl::GetZoneDataVisitorImpl(std::ostream& out, ZoneData* zone_data)
    : out_{out},
      zone_data_{zone_data},
      origin_{},
      soa_id_{},
      insert_horizontal_space_{true}
{ }

std::uint32_t GetZoneDataVisitorImpl::get_soa_id() const noexcept
{
    return soa_id_;
}

void GetZoneDataVisitorImpl::operator()(Common common)
{
    if (common.ttl != boost::none)
    {
        out_ << "$TTL " << *common.ttl << " ;default TTL for all records in zone\n\n";
    }
    if (zone_data_ != nullptr)
    {
        zone_data_->ttl = common.ttl;
    }
    origin_ = std::move(common.soa.zone);
    Soa::RData& r_data = common.soa.r_data;
    soa_id_ = r_data.serial;
    out_ << origin_ << "\tIN\tSOA\t" << r_data.m_name << "\t" << r_data.r_name << " ("
         << r_data.serial << " " << r_data.refresh << " " << r_data.retry << " " << r_data.expire << " " << r_data.minimum << ")\n";
    if (zone_data_ != nullptr)
    {
        if (!zone_data_->origin.empty())
        {
            struct TooManySoaRecords : std::runtime_error
            {
                TooManySoaRecords() : std::runtime_error{std::string{}} { }
                const char* what() const noexcept override { return "too many SOA records"; }
            };
            throw TooManySoaRecords{};
        }
        zone_data_->origin = origin_;
        zone_data_->soa_record = ZoneData::SoaRecord
        {
            std::move(r_data.m_name),
            std::move(r_data.r_name),
            r_data.serial,
            r_data.refresh,
            r_data.retry,
            r_data.expire,
            r_data.minimum
        };
    }
}

std::ostream& GetZoneDataVisitorImpl::data_section(const std::string& owner, const std::string& record_type)
{
    if (insert_horizontal_space_)
    {
        out_ << "\n";
        insert_horizontal_space_ = false;
    }
    return zone_section(owner, record_type);
}

std::ostream& GetZoneDataVisitorImpl::zone_section(const std::string& owner, const std::string& record_type)
{
    return out_ << owner << "\tIN\t" << record_type << "\t";
}

void GetZoneDataVisitorImpl::operator()(Record record)
{
    std::for_each(begin(record.ns), end(record.ns), [&](const NsRData& r_data)
    {
        const bool are_zone_nameservers = record.owner == origin_;
        if (are_zone_nameservers)
        {
            zone_section("", "NS") << r_data.nameserver << "\n";
        }
        else
        {
            data_section(record.owner, "NS") << r_data.nameserver << "\n";
        }
    });
    std::for_each(begin(record.ds), end(record.ds), [&](const DsRData& r_data)
    {
        data_section(record.owner, "DS") << r_data.key_tag << " " << int{r_data.algorithm} << " "
                                         << int{r_data.digest_type} << " " << r_data.digest << "\n";
    });
    std::for_each(begin(record.a), end(record.a), [&](const ARData& r_data)
    {
        data_section(record.owner, "A") << r_data.address.to_string() << "\n";
    });
    std::for_each(begin(record.aaaa), end(record.aaaa), [&](const AaaaRData& r_data)
    {
        data_section(record.owner, "AAAA") << r_data.address.to_string() << "\n";
    });
    if (zone_data_ != nullptr)
    {
        if (!record.ns.empty() || !record.ds.empty())
        {
            ZoneData::DomainData& domain_data = zone_data_->domains[record.owner];
            std::for_each(begin(record.ns), end(record.ns), [&](const NsRData& r_data)
            {
                const bool unique_ns = domain_data.nameservers.insert(r_data.nameserver).second;
                if (!unique_ns)
                {
                    struct MultipleNameservers : std::runtime_error
                    {
                        MultipleNameservers() : std::runtime_error{std::string{}} { }
                        const char* what() const noexcept override { return "nameservers must be unique"; }
                    };
                    throw MultipleNameservers{};
                }
            });
            std::for_each(begin(record.ds), end(record.ds), [&](const DsRData& r_data)
            {
                domain_data.ds_records.insert(ZoneData::DomainData::DsRecord{
                        r_data.key_tag,
                        r_data.algorithm,
                        r_data.digest_type,
                        r_data.digest});
            });
        }
        if (!record.a.empty())
        {
            std::set<boost::asio::ip::address_v4>& addresses = zone_data_->a_records[record.owner];
            std::for_each(begin(record.a), end(record.a), [&](const ARData& r_data)
            {
                const bool a_record_is_unique = addresses.insert(r_data.address).second;
                if (!a_record_is_unique)
                {
                    struct MultipleARecord : std::runtime_error
                    {
                        MultipleARecord() : std::runtime_error{std::string{}} { }
                        const char* what() const noexcept override { return "A records must be unique"; }
                    };
                    throw MultipleARecord{};
                }
            });
        }
        if (!record.aaaa.empty())
        {
            std::set<boost::asio::ip::address_v6>& addresses = zone_data_->aaaa_records[record.owner];
            std::for_each(begin(record.aaaa), end(record.aaaa), [&](const AaaaRData& r_data)
            {
                const bool aaaa_record_is_unique = addresses.insert(r_data.address).second;
                if (!aaaa_record_is_unique)
                {
                    struct MultipleAaaaRecord : std::runtime_error
                    {
                        MultipleAaaaRecord() : std::runtime_error{std::string{}} { }
                        const char* what() const noexcept override { return "AAAA records must be unique"; }
                    };
                    throw MultipleAaaaRecord{};
                }
            });
        }
    }
}

void GetZoneDataVisitorImpl::operator()(NotFound exception)
{
    throw std::runtime_error{exception.message};
}

void GetZoneDataVisitorImpl::operator()(Exception exception)
{
    throw std::runtime_error{exception.message};
}

}//namespace Fred::Zone::Client::{anonymous}

}//namespace Fred::Zone::Client
}//namespace Fred::Zone
}//namespace Fred
