/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "client/src/hook_exec.hh"

#include "liblog/liblog.hh"

#include <boost/filesystem.hpp>

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <cstring>
#include <functional>
#include <memory>
#include <regex>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

namespace Fred {
namespace Zone {
namespace Client {

namespace {

int exec(std::string cmd, const std::vector<std::string>& args);

}//namespace Fred::Zone::Client::{anonymous}

void hook_exec(const std::string& zone, const std::string& zone_file_name, const std::string& backup_zone_file_name, const Cfg::Options::Zone& zone_options)
{
    constexpr auto regex_options = std::regex_constants::extended | std::regex_constants::optimize;
    static const auto regex = std::regex{"(^|[^[:alnum:]])(\\$(file|zone|backup))($|[^[:alnum:]])", regex_options};
    std::vector<std::string> args;
    std::for_each(begin(zone_options.post_hook_arguments), end(zone_options.post_hook_arguments), [&](const std::string& argument)
    {
        std::smatch match_iter{};
        auto arg_begin = cbegin(argument);
        const auto arg_end = cend(argument);
        auto transformed_argument = std::string{};
        while (arg_begin != arg_end)
        {
            if (std::regex_search(arg_begin, arg_end, match_iter, regex))
            {
                transformed_argument.append(arg_begin, match_iter[2].first);
                if (match_iter[3] == "file")
                {
                    transformed_argument.append(zone_file_name);
                }
                else if (match_iter[3] == "zone")
                {
                    transformed_argument.append(zone);
                }
                else
                {
                    transformed_argument.append(backup_zone_file_name);
                }
                arg_begin = match_iter[2].second;
            }
            else
            {
                transformed_argument.append(arg_begin, arg_end);
                break;
            }
        }
        args.emplace_back(std::move(transformed_argument));
    });
    const auto hook_exit_code = exec(*zone_options.post_hook_command, args);
    if (hook_exit_code == EXIT_SUCCESS)
    {
        return;
    }
    throw std::runtime_error{"hook failed"};
}

namespace {

class ExecArgs
{
public:
    explicit ExecArgs(std::string cmd, const std::vector<std::string>& args)
        : cmd_{std::move(cmd)},
          args_content_{nullptr},
          argv_{init_argv(cmd_, args, args_content_)}
    { }
    const char* cmd() const noexcept
    {
        return cmd_.c_str();
    }
    char* const* argv() const noexcept
    {
        return argv_.data();
    }
private:
    static std::vector<char*> init_argv(
            const std::string& cmd,
            const std::vector<std::string>& args,
            std::unique_ptr<char[]>& args_content)
    {
        const auto cmd_file_name_start = [&]() -> std::size_t
        {
            const auto idx = cmd.rfind('/');
            if (idx == std::string::npos)
            {
                return 0;
            }
            return idx + 1;
        }();
        std::vector<char*> argv;
        argv.reserve(args.size() + 2);
        const auto args_content_size = std::size_t{cmd.length() - cmd_file_name_start + 1 +
                [&]()
                {
                    auto size = args.size();
                    std::for_each(begin(args), end(args), [&](const std::string& argument)
                    {
                        size += argument.length();
                    });
                    return size;
                }()};
        static const auto append = [](char*& dst, const char* src, std::size_t src_length)
        {
            char* const result = std::strcpy(dst, src);
            dst += src_length + 1;
            return result;
        };
        args_content = std::unique_ptr<char[]>(new char[args_content_size]);
        char* arg_ptr = args_content.get();
        argv.push_back(append(arg_ptr, cmd.c_str() + cmd_file_name_start, cmd.length() - cmd_file_name_start));
        std::transform(begin(args), end(args), back_inserter(argv), [&](const std::string& argument)
        {
            return append(arg_ptr, argument.c_str(), argument.length());
        });
        argv.push_back(nullptr);
        return argv;
    }
    std::string cmd_;
    std::unique_ptr<char[]> args_content_;
    std::vector<char*> argv_;
};

auto fork(std::function<int()> child)
{
    const auto fork_result = ::fork();
    const bool fork_failed = fork_result == -1;
    if (fork_failed)
    {
        const auto c_errno = errno;
        LIBLOG_ERROR("fork failed: {}", std::strerror(c_errno));
        throw std::runtime_error{std::string{"fork failed: "} + std::strerror(c_errno)};
    }
    const bool im_child = fork_result == 0;
    if (!im_child)
    {
        return fork_result;
    }
    try
    {
        ::_exit(child());
    }
    catch (...)
    {
        ::_exit(EXIT_FAILURE);
    }
}

int wait(::pid_t child)
{
    int process_status;
    while (true)
    {
        const auto wait_result = ::waitpid(child, &process_status, 0);
        const bool wait_failed = wait_result == -1;
        if (wait_failed)
        {
            const auto c_errno = errno;
            if (c_errno == EINTR)
            {
                LIBLOG_INFO("waiting on child {} interrupted by a signal", child);
                continue;
            }
            LIBLOG_ERROR("waiting on child {} failed: ", child, std::strerror(c_errno));
            throw std::runtime_error{std::strerror(c_errno)};
        }
        if (WIFEXITED(process_status))
        {
            const auto exit_status = WEXITSTATUS(process_status);
            if (exit_status == EXIT_SUCCESS)
            {
                LIBLOG_DEBUG("child {} exited successfully", child);
            }
            else
            {
                LIBLOG_INFO("child {} exited with error status {}", child, exit_status);
            }
            return exit_status;
        }
        if (WIFSIGNALED(process_status))
        {
            const auto signal_number = WTERMSIG(process_status);
            LIBLOG_INFO("child {} terminated by signal {}", child, signal_number);
            throw std::runtime_error{"child terminated by a signal"};
        }
        if (WIFSTOPPED(process_status))
        {
            LIBLOG_INFO("child {} stopped", child);
            continue;
        }
        if (WIFCONTINUED(process_status))
        {
            LIBLOG_INFO("child {} continued", child);
            continue;
        }
        LIBLOG_WARNING("child {} changed state without a known reason", child);
    }
}

int exec(std::string cmd, const std::vector<std::string>& args)
{
    const auto child_pid = fork([&]()
    {
        const ExecArgs command{std::move(cmd), args};
        ::execv(command.cmd(), command.argv());
        const auto c_errno = errno;
        LIBLOG_ERROR("exec {} failed: {}", command.cmd(), std::strerror(c_errno));
        return EXIT_FAILURE;
    });
    return wait(child_pid);
}

}//namespace Fred::Zone::Client::{anonymous}

}//namespace Fred::Zone::Client
}//namespace Fred::Zone
}//namespace Fred
