/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "client/src/unwrap_get_zone_data.hh"

#include "liblog/liblog.hh"

#include "fred_api/zone/service_zone_grpc.grpc.pb.h"

#include <grpc++/grpc++.h>

#include <algorithm>
#include <type_traits>

namespace Fred {
namespace Zone {
namespace Client {

namespace {

GetZoneDataVisitor::Soa::RData unwrap(const Api::SoaRecord::Rdata& src)
{
    return GetZoneDataVisitor::Soa::RData{
            src.mname().absolute_fqdn(),
            src.rname().absolute_fqdn(),
            src.serial(),
            src.refresh(),
            src.retry(),
            src.expire(),
            src.minimum()};
}

GetZoneDataVisitor::Soa unwrap(const Api::SoaRecord& src)
{
    return GetZoneDataVisitor::Soa{
            src.zone().absolute_fqdn(),
            unwrap(src.rdata())};
}

GetZoneDataVisitor::Common unwrap_common(const Api::ZoneDataReply::Data& src)
{
    if (src.OptionalTtl_case() == Api::ZoneDataReply::Data::kTtl)
    {
        return GetZoneDataVisitor::Common{src.ttl(), unwrap(src.soa_record())};
    }
    return GetZoneDataVisitor::Common{boost::none, unwrap(src.soa_record())};
}

#if 106600 <= BOOST_VERSION
decltype(auto) make_address_v4(const std::string& addr)
{
    return boost::asio::ip::make_address_v4(addr);
}

decltype(auto) make_address_v6(const std::string& addr)
{
    return boost::asio::ip::make_address_v6(addr);
}
#else
decltype(auto) make_address_v4(const std::string& addr)
{
    return boost::asio::ip::address_v4::from_string(addr);
}

decltype(auto) make_address_v6(const std::string& addr)
{
    return boost::asio::ip::address_v6::from_string(addr);
}
#endif

GetZoneDataVisitor::DsRData unwrap(const Api::DomainRecord::DsRdata& src)
{
    return GetZoneDataVisitor::DsRData{
            static_cast<std::uint16_t>(src.key_tag()),
            static_cast<std::uint8_t>(src.algorithm()),
            static_cast<std::uint8_t>(src.digest_type()),
            src.digest().hexa_value()};
}

GetZoneDataVisitor::NsRData unwrap(const Api::DomainRecord::NsRdata& src)
{
    return GetZoneDataVisitor::NsRData{src.nameserver().absolute_fqdn()};
}

GetZoneDataVisitor::ARData unwrap(const Api::DomainRecord::ARdata& src)
{
    return GetZoneDataVisitor::ARData{make_address_v4(src.ipv4().address())};
}

GetZoneDataVisitor::AaaaRData unwrap(const Api::DomainRecord::AaaaRdata& src)
{
    return GetZoneDataVisitor::AaaaRData{make_address_v6(src.ipv6().address())};
}

template <typename Type, template <typename> class Vector>
std::enable_if_t<std::is_arithmetic<decltype(std::declval<const Vector<Type>&>().size())>::value &&
                 std::is_same<decltype(*std::declval<const Vector<Type>&>().begin()), const Type&>::value,
                 std::vector<decltype(unwrap(std::declval<const Type&>()))>> unwrap(const Vector<Type>& src)
{
    using Dst = decltype(unwrap(std::declval<const Type&>()));
    std::vector<Dst> dst;
    dst.reserve(src.size());
    std::transform(src.begin(), src.end(), std::back_inserter(dst), [](const Type& item)
    {
        return unwrap(item);
    });
    return dst;
}

}//namespace Fred::Zone::Client::{anonymous}

void unwrap_get_zone_data(const Cfg::Options::ZoneService& zone_service_options, const std::string& zone, GetZoneDataVisitor& visitor)
{
    auto channel = grpc::CreateChannel(zone_service_options.connect_to, grpc::InsecureChannelCredentials());
    ::grpc::ClientContext ctx;
    Api::ZoneDataRequest request;
    *(request.mutable_zone()->mutable_fqdn()) = zone;
    auto reader = Api::Zone::NewStub(channel)->get_zone_data(&ctx, request);
    LIBLOG_DEBUG("have a reader");
    Api::ZoneDataReply reply;
    auto read_cnt = 0;
    while (reader->Read(&reply))
    {
        if (reply.data().has_soa_record())
        {
            visitor(unwrap_common(reply.data()));
        }
        std::for_each(std::begin(reply.data().records()), std::end(reply.data().records()), [&](const Api::DomainRecord& src_record)
        {
            GetZoneDataVisitor::Record dst_record;
            dst_record.owner = src_record.owner().absolute_fqdn();
            std::for_each(src_record.ns_records().begin(), src_record.ns_records().end(), [&](const Api::DomainRecord::NsRdata& src_rdata)
            {
                dst_record.ns.emplace_back(unwrap(src_rdata));
            });
            std::for_each(src_record.ds_records().begin(), src_record.ds_records().end(), [&](const Api::DomainRecord::DsRdata& src_rdata)
            {
                dst_record.ds.emplace_back(unwrap(src_rdata));
            });
            std::for_each(src_record.a_records().begin(), src_record.a_records().end(), [&](const Api::DomainRecord::ARdata& src_rdata)
            {
                dst_record.a.emplace_back(unwrap(src_rdata));
            });
            std::for_each(src_record.aaaa_records().begin(), src_record.aaaa_records().end(), [&](const Api::DomainRecord::AaaaRdata& src_rdata)
            {
                dst_record.aaaa.emplace_back(unwrap(src_rdata));
            });
            visitor(std::move(dst_record));
        });
        if (read_cnt == 0)
        {
            LIBLOG_DEBUG("first read done");
        }
        ++read_cnt;
    }
    LIBLOG_DEBUG("last ({}.) read done", read_cnt);
    const auto status = reader->Finish();
    if (status.ok())
    {
        return;
    }
    switch (status.error_code())
    {
        case grpc::StatusCode::NOT_FOUND:
            LIBLOG_INFO("get_zone_data failure: {}", status.error_message());
            visitor(GetZoneDataVisitor::NotFound{status.error_message()});
            break;
        default:
            LIBLOG_ERROR("get_zone_data failure: {}", status.error_message());
            visitor(GetZoneDataVisitor::Exception{status.error_message()});
            break;
    }
    struct GetZoneDataFailed : std::runtime_error
    {
        GetZoneDataFailed() : std::runtime_error{std::string{}} { }
        const char* what() const noexcept override { return "get_zone_data failed"; }
    };
    throw GetZoneDataFailed{};
}

}//namespace Fred::Zone::Client
}//namespace Fred::Zone
}//namespace Fred
