/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/iface/tools/common_types_wrappers.hh"
#include "src/iface/tools/generic_wrappers.hh"

#include <algorithm>

namespace Fred {
namespace Zone {
namespace Iface {
namespace Tools {

void wrap(const Impl::Lib::ZoneData::SoaRecord::RData& src, Api::SoaRecord::Rdata* dst)
{
    wrap(*src.m_name, dst->mutable_mname());
    wrap(*src.r_name, dst->mutable_rname());
    dst->set_serial(*src.serial);
    dst->set_refresh(*src.refresh);
    dst->set_retry(*src.retry);
    dst->set_expire(*src.expire);
    dst->set_minimum(*src.minimum);
}

void wrap(const Impl::Lib::ZoneData::Common& src, Api::ZoneDataReply::Data* dst)
{
    if (has_value(src.ttl))
    {
        dst->set_ttl(**src.ttl);
    }
    auto* const soa_record = dst->mutable_soa_record();
    wrap(make_absolute(Impl::Lib::Fqdn{*src.soa_record.zone}), soa_record->mutable_zone());
    wrap(src.soa_record.r_data, soa_record->mutable_rdata());
}

void wrap(const Impl::Lib::AbsoluteFqdn& src, Api::AbsoluteDomainName* dst)
{
    wrap(*src.get_fqdn(), dst->mutable_absolute_fqdn());
}

namespace {

void wrap_glue_records(
        const Impl::Lib::AbsoluteFqdn& nameserver,
        const Impl::Lib::ZoneData::Nameserver::GlueRecords& glue_records,
        Api::DomainRecord* record)
{
    wrap(nameserver, record->mutable_owner());
    if (!glue_records.a_records.empty())
    {
        auto* const r_data = record->mutable_a_records();
        r_data->Reserve(glue_records.a_records.size());
        std::for_each(
                    begin(glue_records.a_records),
                    end(glue_records.a_records),
                    [&](const boost::asio::ip::address_v4& address)
        {
            wrap(address, r_data->Add());
        });
    }
    if (!glue_records.aaaa_records.empty())
    {
        auto* const r_data = record->mutable_aaaa_records();
        r_data->Reserve(glue_records.aaaa_records.size());
        std::for_each(
                    begin(glue_records.aaaa_records),
                    end(glue_records.aaaa_records),
                    [&](const boost::asio::ip::address_v6& address)
        {
            wrap(address, r_data->Add());
        });
    }
}

void wrap_nameservers(
        const std::vector<Impl::Lib::ZoneData::Nameserver>& src,
        ::google::protobuf::RepeatedPtrField<Api::DomainRecord>* dst,
        ::google::protobuf::RepeatedPtrField<Api::DomainRecord::NsRdata>* r_data)
{
    std::for_each(begin(src), end(src), [&](const Impl::Lib::ZoneData::Nameserver& nameserver)
    {
        wrap(nameserver.get_nameserver_hostname(), r_data->Add()->mutable_nameserver());
        if (nameserver.are_glue_records_permitted() && (!nameserver.get_glue_records().a_records.empty() ||
                                                        !nameserver.get_glue_records().aaaa_records.empty()))
        {
            wrap_glue_records(nameserver.get_nameserver_hostname(), nameserver.get_glue_records(), dst->Add());
        }
    });
}

}//namespace Fred::Zone::Iface::Tools::{anonymous}

void wrap(const Impl::Lib::ZoneData::Domain& src, ::google::protobuf::RepeatedPtrField<Api::DomainRecord>* dst)
{
    if (src.get_nameservers().empty() && src.get_ds_r_data().empty())
    {
        return;
    }
    Api::DomainRecord* const record = dst->Add();
    wrap(src.get_domain_name(), record->mutable_owner());
    if (!src.get_ds_r_data().empty())
    {
        ::google::protobuf::RepeatedPtrField<Api::DomainRecord::DsRdata>* const r_data = record->mutable_ds_records();
        r_data->Reserve(src.get_ds_r_data().size());
        std::for_each(begin(src.get_ds_r_data()), end(src.get_ds_r_data()), [&](const Impl::Lib::ZoneData::DsRData& ds_r_data)
        {
            wrap(ds_r_data, r_data->Add());
        });
    }
    if (!src.get_nameservers().empty())
    {
        ::google::protobuf::RepeatedPtrField<Api::DomainRecord::NsRdata>* const r_data = record->mutable_ns_records();
        r_data->Reserve(src.get_nameservers().size());
        wrap_nameservers(src.get_nameservers(), dst, r_data);
    }
}

void wrap(const boost::asio::ip::address_v4& src, Api::DomainRecord::ARdata* dst)
{
    wrap(src, dst->mutable_ipv4());
}

void wrap(const boost::asio::ip::address_v6& src, Api::DomainRecord::AaaaRdata* dst)
{
    wrap(src, dst->mutable_ipv6());
}

void wrap(const boost::asio::ip::address_v4& src, Api::IPv4Address* dst)
{
    wrap(src.to_string(), dst->mutable_address());
}

void wrap(const boost::asio::ip::address_v6& src, Api::IPv6Address* dst)
{
    wrap(src.to_string(), dst->mutable_address());
}

void wrap(const Impl::Lib::ZoneData::DsRData& src, Api::DomainRecord::DsRdata* dst)
{
    dst->set_key_tag(*src.get_key_tag());
    dst->set_algorithm(*src.get_algorithm());
    dst->set_digest_type(src.get_digest_type().get_value());
    wrap(src.get_digest(), dst->mutable_digest());
}

void wrap(const Impl::Lib::ZoneData::DsRData::Digest& src, Api::DomainRecord::DsRdata::Digest* dst)
{
    wrap(*src, dst->mutable_hexa_value());
}

void wrap(const Impl::Lib::ZoneName& src, Api::ZoneName* dst)
{
    dst->set_fqdn(*src);
}

void wrap(const Impl::Lib::ZoneList& src, Api::ZoneListReply* dst)
{
    wrap(src, dst->mutable_data()->mutable_zones());
}

}//namespace Fred::Zone::Iface::Tools
}//namespace Fred::Zone::Iface
}//namespace Fred::Zone
}//namespace Fred
