/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/iface/zone_service.hh"

#include "src/impl/zone_data.hh"
#include "src/impl/zone_list.hh"

#include "src/iface/tools/tools.hh"
#include "src/iface/tools/common_types_unwrappers.hh"
#include "src/iface/tools/common_types_wrappers.hh"
#include "src/iface/tools/generic_wrappers.hh"

#include "liblog/liblog.hh"

#include <algorithm>
#include <cmath>

namespace Fred {
namespace Zone {
namespace Iface {

grpc::Status ZoneService::get_zone_list(
        grpc::ServerContext*,
        const ::google::protobuf::Empty*,
        ::Fred::Zone::Api::ZoneListReply* reply)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "Zone", __func__);
    try
    {
        const auto result = Impl::zone_list();
        Tools::wrap(result, reply);
        return grpc::Status::OK;
    }
    catch (const Impl::Lib::RuntimeError& e)
    {
        return Tools::on_runtime_error(e);
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_database_error(e);
    }
    catch (const LibPg::ConnectFailure& e)
    {
        return Tools::on_database_error(e);
    }
    catch (const LibPg::PgException& e)
    {
        return Tools::on_database_error(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

grpc::Status ZoneService::get_zone_data(
        grpc::ServerContext*,
        const ::Fred::Zone::Api::ZoneDataRequest* api_request,
        grpc::ServerWriter<Api::ZoneDataReply>* writer)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "Zone", __func__);
    try
    {
        const auto zone = Tools::unwrap(api_request->zone());
        const auto zone_data = Impl::get_zone_data(zone);
        Api::ZoneDataReply reply;
        auto* const reply_data = reply.mutable_data();
        Tools::wrap(zone_data.get_common(), reply_data);
        Tools::wrap(zone_data.get_zone_nameservers(), reply_data->mutable_records());
        auto zone_data_iter = zone_data.begin();
        if (zone_data_iter == zone_data.end())
        {
            writer->Write(reply);
        }
        else
        {
            do
            {
                constexpr auto records_chunk_size = 1024;
                auto* const records = reply_data->mutable_records();
                records->Reserve(records_chunk_size);
                while ((reply.data().records().size() < records_chunk_size) &&
                       (zone_data_iter != zone_data.end()))
                {
                    Tools::wrap(*zone_data_iter, records);
                    ++zone_data_iter;
                }
                writer->Write(reply);
                reply_data->clear_ttl();
                reply_data->clear_soa_record();
                reply_data->clear_records();
            } while (zone_data_iter != zone_data.end());
        }
        return grpc::Status::OK;
    }
    catch (const Impl::Lib::NotFound& e)
    {
        return Tools::on_not_found(e);
    }
    catch (const Impl::Lib::RuntimeError& e)
    {
        return Tools::on_runtime_error(e);
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_database_error(e);
    }
    catch (const LibPg::ConnectFailure& e)
    {
        return Tools::on_database_error(e);
    }
    catch (const LibPg::PgException& e)
    {
        return Tools::on_database_error(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

}//namespace Fred::Zone::Iface
}//namespace Fred::Zone
}//namespace Fred
