/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/absolute_fqdn.hh"
#include "src/impl/lib/exception.hh"

#include <utility>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

namespace {

class FqdnView
{
public:
    template <std::size_t n>
    explicit constexpr FqdnView(const char (&str)[n])
        : begin_{str},
          end_{str + n - 1}
    {
        static_assert(0 < n, "string must occupy at least one character");
    }
    explicit constexpr FqdnView(const char* begin, const char* end)
        : begin_{begin},
          end_{end}
    { }
    explicit FqdnView(const Fqdn& fqdn)
        : FqdnView{*fqdn}
    { }
    constexpr FqdnView(const FqdnView&) noexcept = default;
    constexpr FqdnView(FqdnView&&) noexcept = default;
    constexpr FqdnView& operator=(const FqdnView&) noexcept = default;
    constexpr FqdnView& operator=(FqdnView&&) noexcept = default;
    constexpr auto length() const noexcept
    {
        return end_ - begin_;
    }
    constexpr bool is_empty() const noexcept
    {
        return this->length() <= 0;
    }
    constexpr bool is_same(const char* rhs) const noexcept
    {
        for (const char* lhs = begin_; lhs < end_; ++lhs, ++rhs)
        {
            if (*lhs != *rhs)
            {
                return false;
            }
        }
        return true;
    }
    constexpr bool is_subdomain_of(FqdnView domain) const noexcept
    {
        return is_subdomain(*this, std::move(domain));
    }
    constexpr char front() const noexcept
    {
        return *begin_;
    }
    constexpr char back() const noexcept
    {
        return end_[-1];
    }
    constexpr FqdnView shrink_back() const noexcept
    {
        return FqdnView{begin_, end_ - 1};
    }
    template <typename T>
    constexpr FqdnView tail(T length) const noexcept
    {
        return FqdnView{end_ - length, end_};
    }
private:
    explicit FqdnView(const std::string& str)
        : begin_{&*std::begin(str)},
          end_{&*std::end(str)}
    { }
    static constexpr bool is_subdomain(FqdnView subdomain, FqdnView domain) noexcept
    {
        if (subdomain.is_empty())
        {
            return false;
        }
        if (subdomain.back() == '.')
        {
            subdomain = subdomain.shrink_back();
            if (subdomain.is_empty())
            {
                return false;
            }
        }
        if (domain.is_empty())
        {
            return true;
        }
        if (domain.back() == '.')
        {
            domain = domain.shrink_back();
            if (domain.is_empty())
            {
                return true;
            }
        }
        const auto min_length_of_subdomain = domain.length() + 2;
        if (subdomain.length() < min_length_of_subdomain)
        {
            return false;
        }
        if (subdomain.tail(domain.length() + 1).front() != '.')
        {
            return false;
        }
        return subdomain.tail(domain.length()) == domain;
    }
    const char* begin_;
    const char* end_;
    friend constexpr bool operator==(const FqdnView& lhs, const FqdnView& rhs) noexcept
    {
        return (lhs.length() == rhs.length()) && lhs.is_same(rhs.begin_);
    }
    friend constexpr bool operator<(const FqdnView& lhs, const FqdnView& rhs) noexcept
    {
        const char* lhs_top = lhs.begin_;
        const char* rhs_top = rhs.begin_;
        while ((lhs_top != lhs.end_) && (rhs_top != rhs.end_))
        {
            if (*lhs_top < *rhs_top)
            {
                return true;
            }
            if (*rhs_top < *lhs_top)
            {
                return false;
            }
            ++lhs_top;
            ++rhs_top;
        }
        return (lhs_top == lhs.end_) && (rhs_top != rhs.end_);
    }
};

FqdnView make_relative_fqdn(const Fqdn& fqdn)
{
    const auto src = FqdnView{fqdn};
    if (src.is_empty() || (src.back() != '.'))
    {
        return src;
    }
    return src.shrink_back();
}

static_assert(FqdnView{"sub.domain.cz."}.back() == '.', "must be terminated by '.'");

static_assert(FqdnView{"sub.domain.cz"}.is_subdomain_of(FqdnView{"domain.cz"}), "must be subdomain");
static_assert(FqdnView{"sub.domain.cz."}.is_subdomain_of(FqdnView{"domain.cz"}), "must be subdomain");
static_assert(FqdnView{"sub.domain.cz"}.is_subdomain_of(FqdnView{"domain.cz."}), "must be subdomain");
static_assert(FqdnView{"sub.domain.cz."}.is_subdomain_of(FqdnView{"domain.cz."}), "must be subdomain");
static_assert(FqdnView{"sub.domain.cz."}.is_subdomain_of(FqdnView{""}), "must be subdomain");
static_assert(FqdnView{"sub.domain.cz."}.is_subdomain_of(FqdnView{"."}), "must be subdomain");

static_assert(!FqdnView{"domain.cz"}.is_subdomain_of(FqdnView{"domain.cz"}), "can not be subdomain");
static_assert(!FqdnView{".domain.cz"}.is_subdomain_of(FqdnView{"domain.cz"}), "can not be subdomain");
static_assert(!FqdnView{"domain.cz"}.is_subdomain_of(FqdnView{"sub.domain.cz"}), "can not be subdomain");
static_assert(!FqdnView{""}.is_subdomain_of(FqdnView{"domain.cz"}), "can not be subdomain");
static_assert(!FqdnView{""}.is_subdomain_of(FqdnView{""}), "can not be subdomain");
static_assert(!FqdnView{"."}.is_subdomain_of(FqdnView{""}), "can not be subdomain");
static_assert(!FqdnView{""}.is_subdomain_of(FqdnView{"."}), "can not be subdomain");
static_assert(!FqdnView{"."}.is_subdomain_of(FqdnView{"."}), "can not be subdomain");

constexpr bool is_absolute(const FqdnView& fqdn)
{
    return !fqdn.is_empty() && (fqdn.back() == '.');
}

decltype(auto) fqdn_checked_for_length(Fqdn fqdn)
{
    constexpr auto max_fqdn_length = 255;
    if (max_fqdn_length < (make_relative_fqdn(fqdn).length() + 1))
    {
        throw_runtime_error([](){ return "fqdn must be at most 255 characters long"; });
    }
    return fqdn;
}

std::string make_absolute_fqdn(const Fqdn& fqdn)
{
    if (is_absolute(FqdnView{fqdn}))
    {
        return *fqdn;
    }
    return (*fqdn) + ".";
}

}//namespace Fred::Zone::Impl::Lib::{anonymous}

AbsoluteFqdn::AbsoluteFqdn(Fqdn fqdn)
    : fqdn_{fqdn_checked_for_length(std::move(fqdn))}
{ }

Fqdn AbsoluteFqdn::get_fqdn() const
{
    return Fqdn{make_absolute_fqdn(fqdn_)};
}

bool AbsoluteFqdn::is_subdomain_of(const AbsoluteFqdn& domain) const noexcept
{
    return FqdnView{fqdn_}.is_subdomain_of(FqdnView{domain.fqdn_});
}

bool operator<(const AbsoluteFqdn& lhs, const AbsoluteFqdn& rhs) noexcept
{
    return make_relative_fqdn(lhs.fqdn_) < make_relative_fqdn(rhs.fqdn_);
}

AbsoluteFqdn make_absolute(Fqdn fqdn)
{
    return AbsoluteFqdn{std::move(fqdn)};
}

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred
