/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef BASE64_HH_3890307E64115F7F562CCE6F0525882E//date "+%s.%N"|md5sum|tr "[a-f]" "[A-F]"
#define BASE64_HH_3890307E64115F7F562CCE6F0525882E

#include "src/impl/lib/detail/base64.hh"
#include "src/impl/lib/strong_type.hh"

#include <vector>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

using Detail::Base64DecoderFailure;

template <typename InputIterator, typename OutputContainer>
void decode_base64(InputIterator begin, InputIterator end, std::back_insert_iterator<OutputContainer> output)
{
    static_assert(std::is_same<std::decay_t<decltype(*begin)>, char>::value, "input iterator must point at char type");
    using OutputValueType = decltype(*std::begin(std::declval<OutputContainer>()));
    static_assert(std::is_same<std::decay_t<OutputValueType>, std::uint8_t>::value, "output iterator must point at std::uint8_t type");
    const auto base64_data_length = end - begin;
    const bool nothing_to_decode = base64_data_length == 0;
    if (nothing_to_decode)
    {
        return;
    }
    constexpr auto base64_data_alignment = 4;
    const bool is_correctly_aligned = (base64_data_length % base64_data_alignment) == 0;
    if (!is_correctly_aligned)
    {
        struct IncorrectBase64DataLength : Base64DecoderFailure
        {
            const char* what() const noexcept override { return "incorrect length of Base64 encoded data"; }
        };
        throw IncorrectBase64DataLength{};
    }
    const auto number_of_chunks = base64_data_length / base64_data_alignment;
    for (std::decay_t<decltype(number_of_chunks)> chunk_idx = 0; (chunk_idx + 1) < number_of_chunks; ++chunk_idx)
    {
        const auto b0 = Detail::decode_base64_char(*begin);
        ++begin;
        const auto b1 = Detail::decode_base64_char(*begin);
        ++begin;
        const auto b2 = Detail::decode_base64_char(*begin);
        ++begin;
        const auto b3 = Detail::decode_base64_char(*begin);
        ++begin;
        Detail::join_6bit_values(b0, b1, b2, b3, output);
    }

    const auto b0 = Detail::decode_base64_char(*begin);
    ++begin;
    const auto b1 = Detail::decode_base64_char(*begin);
    ++begin;
    const auto c2 = *begin;
    ++begin;
    const auto c3 = *begin;
    constexpr char align_character = '=';
    if (c2 != align_character)
    {
        const auto b2 = Detail::decode_base64_char(c2);
        if (c3 != align_character)
        {
            const auto b3 = Detail::decode_base64_char(c3);
            Detail::join_6bit_values(b0, b1, b2, b3, output);
        }
        else
        {
            Detail::join_6bit_values(b0, b1, b2, output);
        }
    }
    else
    {
        Detail::join_6bit_values(b0, b1, output);
    }
}

template <typename InputContainer>
std::vector<std::uint8_t> decode_base64(const InputContainer& base64_encoded_data)
{
    std::vector<std::uint8_t> binary_data;
    binary_data.reserve(Detail::estimate_base64_decoded_data_length(std::begin(base64_encoded_data), std::end(base64_encoded_data)));
    decode_base64(std::begin(base64_encoded_data), std::end(base64_encoded_data), back_inserter(binary_data));
    return binary_data;
}

using Base64EncodedData = StrongType<std::string, struct Base64EncodedDataTag_>;

std::vector<std::uint8_t> decode(const Base64EncodedData& base64_encoded_data);

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred

#endif//BASE64_HH_3890307E64115F7F562CCE6F0525882E
