/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef BASE64_HH_CAD7D922D7BB46F3D532FEB5C7A6093F//date "+%s.%N"|md5sum|tr "[a-f]" "[A-F]"
#define BASE64_HH_CAD7D922D7BB46F3D532FEB5C7A6093F

#include "src/impl/lib/exception.hh"

#include <cstdint>
#include <type_traits>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {
namespace Detail {

template <char first, char last> struct CharInterval;

template <typename...> struct Decode;

template <char first, char last, typename ...Next>
struct Decode<CharInterval<first, last>, Next...>
{
    static_assert(first <= last, "first can not be greater than last");
    static constexpr auto interval_length = std::uint8_t{last - first + 1};
    template <typename C>
    constexpr std::uint8_t operator()(C c, std::uint8_t offset = std::uint8_t{0}) const
    {
        static_assert(std::is_same<std::decay_t<C>, char>::value, "C must be char type");
        return (first <= c) && (c <= last) ? static_cast<std::uint8_t>(c - first + offset)
                                           : Decode<Next...>{}(c, static_cast<std::uint8_t>(offset + interval_length));
    }
    static constexpr auto length()
    {
        return interval_length + Decode<Next...>::length();
    }
};

template <>
struct Decode<>
{
    template <typename C>
    constexpr std::uint8_t operator()(C, std::uint8_t = std::uint8_t{0}) const
    {
        static_assert(std::is_same<std::decay_t<C>, char>::value, "C must be char type");
        constexpr auto invalid_value = std::uint8_t{0xff};
        return invalid_value;
    }
    static constexpr auto length()
    {
        return 0;
    }
};

using Base64Decoder = Decode<CharInterval<'A', 'Z'>,
                             CharInterval<'a', 'z'>,
                             CharInterval<'0', '9'>,
                             CharInterval<'+', '+'>,
                             CharInterval<'/', '/'>>;

template <typename C>
constexpr std::enable_if_t<std::is_same<C, char>::value, std::uint8_t> decode_base64_char(C c)
{
    return std::uint8_t{Base64Decoder{}(c)};
}

struct Base64DecoderFailure : RuntimeError { };
struct InvalidBase64Character : Base64DecoderFailure { };
struct UnexpectedBase64Character : Base64DecoderFailure { };

[[noreturn]] void throw_invalid_base64_character();
[[noreturn]] void throw_unexpected_base64_character();

template <typename B>
void check_6bit_length(B b)
{
    if ((b & 0xc0) != 0x00)
    {
        throw_invalid_base64_character();
    }
}

constexpr auto chain_6bit_values_into_24bits(std::uint8_t b0, std::uint8_t b1, std::uint8_t b2 = 0, std::uint8_t b3 = 0)
{
    return std::uint32_t{(std::uint32_t{b0} << 18) | (std::uint32_t{b1} << 12) | (std::uint32_t{b2} << 6) | std::uint32_t{b3}};
}

template <typename OutputContainer>
void join_6bit_values(std::uint8_t b0, std::uint8_t b1, std::uint8_t b2, std::uint8_t b3, std::back_insert_iterator<OutputContainer>& output)
{
    using OutputValueType = decltype(*std::begin(std::declval<OutputContainer>()));
    static_assert(std::is_same<std::decay_t<OutputValueType>, std::uint8_t>::value, "output iterator must point at std::uint8_t type");
    check_6bit_length(b0 | b1 | b2 | b3);
    const auto sum = chain_6bit_values_into_24bits(b0, b1, b2, b3);
    output = static_cast<std::uint8_t>((sum >> 16) & 0xff);
    output = static_cast<std::uint8_t>((sum >> 8) & 0xff);
    output = static_cast<std::uint8_t>(sum & 0xff);
}

template <typename OutputContainer>
void join_6bit_values(std::uint8_t b0, std::uint8_t b1, std::uint8_t b2, std::back_insert_iterator<OutputContainer>& output)
{
    using OutputValueType = decltype(*std::begin(std::declval<OutputContainer>()));
    static_assert(std::is_same<std::decay_t<OutputValueType>, std::uint8_t>::value, "output iterator must point at std::uint8_t type");
    check_6bit_length(b0 | b1 | b2);
    const auto sum = chain_6bit_values_into_24bits(b0, b1, b2);
    const bool unused_bits_are_zero = (sum & 0xff) == 0x00;
    if (!unused_bits_are_zero)
    {
        throw_unexpected_base64_character();
    }
    output = static_cast<std::uint8_t>((sum >> 16) & 0xff);
    output = static_cast<std::uint8_t>((sum >> 8) & 0xff);
}

template <typename OutputContainer>
void join_6bit_values(std::uint8_t b0, std::uint8_t b1, std::back_insert_iterator<OutputContainer>& output)
{
    using OutputValueType = decltype(*std::begin(std::declval<OutputContainer>()));
    static_assert(std::is_same<std::decay_t<OutputValueType>, std::uint8_t>::value, "output iterator must point at std::uint8_t type");
    check_6bit_length(b0 | b1);
    const auto sum = chain_6bit_values_into_24bits(b0, b1);
    const bool unused_bits_are_zero = (sum & 0xffff) == 0x00;
    if (!unused_bits_are_zero)
    {
        throw_unexpected_base64_character();
    }
    output = static_cast<std::uint8_t>((sum >> 16) & 0xff);
}

template <typename InputIterator>
constexpr auto estimate_base64_decoded_data_length(InputIterator begin, InputIterator end) noexcept
{
    const auto base64_data_length = end - begin;
    const bool nothing_to_decode = base64_data_length == 0;
    if (nothing_to_decode)
    {
        return std::int64_t{0};
    }
    constexpr auto base64_data_chunk_size = 4;
    constexpr auto base64_data_chunk_decoded_size = 3;
    const bool is_correctly_aligned = (base64_data_length % base64_data_chunk_size) == 0;
    if (!is_correctly_aligned)
    {
        return std::int64_t{0};
    }
    constexpr char align_character = '=';
    const auto full_length = std::int64_t{(base64_data_chunk_decoded_size * base64_data_length) / base64_data_chunk_size};
    if (*(begin + base64_data_length - 2) == align_character)
    {
        return std::int64_t{full_length - 2};
    }
    if (*(begin + base64_data_length - 1) == align_character)
    {
        return std::int64_t{full_length - 1};
    }
    return full_length;
}

}//namespace Fred::Zone::Impl::Lib::Detail
}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred

#endif//BASE64_HH_CAD7D922D7BB46F3D532FEB5C7A6093F
