/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/ds_record.hh"
#include "src/impl/lib/exception.hh"
#include "src/impl/lib/hash.hh"
#include "src/impl/lib/wire_format.hh"

#include <algorithm>
#include <type_traits>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

namespace {

DsRecord::KeyTag compute_key_tag_by_simple_algorithm(const DnsKey& dns_key)
{
    if (dns_key.public_key().size() < 4)
    {
        throw_runtime_error([](){ return "public key too short (minimum is 4 bytes)"; });
    }
    return DsRecord::KeyTag{static_cast<DsRecord::KeyTag::ValueType>(rbegin(dns_key.public_key())[4] * 0x100 + rbegin(dns_key.public_key())[3])};
}

DsRecord::KeyTag compute_key_tag_by_standard_algorithm(const DnsKey& dns_key)
{
    std::uint32_t sum = (*dns_key.flags()) + (*dns_key.protocol()) * 0x100 + (*dns_key.algorithm());
    bool is_odd = false;
    std::for_each(
                std::begin(dns_key.public_key()),
                std::end(dns_key.public_key()),
                [&](auto byte)
                {
                    static_assert(std::is_same<decltype(byte), std::uint8_t>::value, "byte must be of std::uint8_t type");
                    if (is_odd)
                    {
                        sum += byte;
                    }
                    else
                    {
                        sum += byte * 0x100;
                    }
                    is_odd = !is_odd;
                });
    sum += (sum / 0x10000) % 0x10000;
    return DsRecord::KeyTag{static_cast<DsRecord::KeyTag::ValueType>(sum % 0x10000)};
}

template <HashFunction fnc>
std::string hash_as_string(const std::vector<std::uint8_t>& bytes)
{
    return to_string(compute_hash<fnc>(bytes.data(), bytes.size()));
}

std::uint8_t make_digest_type_value(const std::string& str)
{
    if (str == "sha1")
    {
        return std::uint8_t{1};
    }
    if (str == "sha256")
    {
        return std::uint8_t{2};
    }
    if (str == "sha384")
    {
        return std::uint8_t{4};
    }
    throw_runtime_error([]() { return "unsupported digest type (must be one of sha1, sha256, sha384)"; });
}

std::function<std::string(const std::vector<std::uint8_t>&)> get_hash_function(DsRecord::DigestType digest_type)
{
    switch (digest_type.get_value())
    {
        case std::uint8_t{1}:
            return hash_as_string<HashFunction::sha1>;
        case std::uint8_t{2}:
            return hash_as_string<HashFunction::sha256>;
        case std::uint8_t{4}:
            return hash_as_string<HashFunction::sha384>;
    }
    throw_runtime_error([]() { return "unsupported digest type (must be one of 1, 2, 4)"; });
}

}//namespace Fred::Zone::Impl::Lib::{anonymous}

DsRecord::DigestType::DigestType(const std::string& digest_type)
    : value_{make_digest_type_value(digest_type)}
{ }

std::uint8_t DsRecord::DigestType::get_value() const noexcept
{
    return value_;
}

bool operator<(DsRecord::DigestType lhs, DsRecord::DigestType rhs) noexcept
{
    return lhs.value_ < rhs.value_;
}

DsRecord::KeyTag DsRecord::make_key_tag(const DnsKey& dns_key)
{
    constexpr auto rsa_md5_algorithm = std::uint8_t{1};
    switch (*dns_key.algorithm())
    {
        case rsa_md5_algorithm:
            return compute_key_tag_by_simple_algorithm(dns_key);
        default:
            return compute_key_tag_by_standard_algorithm(dns_key);
    }
}

DsRecord::Digest DsRecord::make_digest(const AbsoluteFqdn& domain, const DnsKey& dns_key, DigestType digest_type)
{
    const auto compute_digest = get_hash_function(digest_type);
    return DsRecord::Digest{compute_digest(WireFormat{}.append(domain).append(dns_key).data())};
}

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred
