/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "src/impl/lib/hash.hh"
#include "src/impl/lib/exception.hh"

#include <openssl/evp.h>

#include <algorithm>
#include <memory>
#include <type_traits>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

namespace  {

void success_required(bool is_success, const char* message)
{
    if (!is_success)
    {
        auto what = std::string{message};
        throw_runtime_error([=]() { return what.c_str(); });
    }
}

template <HashFunction fnc>
const EVP_MD* get_evp_hash_function();

template <>
const EVP_MD* get_evp_hash_function<HashFunction::sha1>()
{
    return EVP_sha1();
}

template <>
const EVP_MD* get_evp_hash_function<HashFunction::sha256>()
{
    return EVP_sha256();
}

template <>
const EVP_MD* get_evp_hash_function<HashFunction::sha384>()
{
    return EVP_sha384();
}

struct EvpMdCtx
{
    struct Creator
    {
        EVP_MD_CTX* operator()() const
        {
#ifdef HAVE_EVP_MD_CTX_NEW
            return EVP_MD_CTX_new();
#elif defined(HAVE_EVP_MD_CTX_CREATE)
            return EVP_MD_CTX_create();
#else
#error "No function to create EVP_MD_CTX*"
#endif
        }
    };
    struct Deleter
    {
        void operator()(EVP_MD_CTX* ctx) const
        {
#ifdef HAVE_EVP_MD_CTX_FREE
            EVP_MD_CTX_free(ctx);
#elif defined(HAVE_EVP_MD_CTX_DESTROY)
            EVP_MD_CTX_destroy(ctx);
#else
#error "No function to free EVP_MD_CTX*"
#endif
        }
    };
};

template <typename T>
std::string& append_hex(T byte, std::string& buffer)
{
    static_assert(std::is_same<decltype(byte), std::uint8_t>::value, "byte must be of std::uint8_t type");
    constexpr char hex[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
    return buffer.append(1, hex[byte >> 4]).append(1, hex[byte & 0x0f]);
}

}//namespace Fred::Zone::Impl::Lib::{anonymous}

template <HashFunction fnc>
Hash<fnc> compute_hash(const void* data, std::size_t data_length)
{
    std::unique_ptr<EVP_MD_CTX, EvpMdCtx::Deleter> ctx{EvpMdCtx::Creator{}()};
    success_required(EVP_DigestInit_ex(ctx.get(), get_evp_hash_function<fnc>(), nullptr) == 1, "EVP_DigestInit_ex failure");
    success_required(EVP_DigestUpdate(ctx.get(), data, data_length) == 1, "EVP_DigestUpdate failure");
    unsigned char md_value[EVP_MAX_MD_SIZE];
    unsigned int md_len;
    success_required(EVP_DigestFinal_ex(ctx.get(), md_value, &md_len) == 1, "EVP_DigestFinal_ex failure");
    return Hash<fnc>{std::vector<std::uint8_t>{md_value, md_value + md_len}};
}

Sha1Hash compute_sha1_hash(const void* data, std::size_t data_length)
{
    return compute_hash<HashFunction::sha1>(data, data_length);
}

Sha256Hash compute_sha256_hash(const void* data, std::size_t data_length)
{
    return compute_hash<HashFunction::sha256>(data, data_length);
}

Sha384Hash compute_sha384_hash(const void* data, std::size_t data_length)
{
    return compute_hash<HashFunction::sha384>(data, data_length);
}

template Hash<HashFunction::sha1> compute_hash(const void*, std::size_t);
template Hash<HashFunction::sha256> compute_hash(const void*, std::size_t);
template Hash<HashFunction::sha384> compute_hash(const void*, std::size_t);

template <HashFunction fnc>
std::string to_string(const Hash<fnc>& data)
{
    std::string result{};
    result.reserve(data->size() * 2);
    std::for_each(std::begin(*data), std::end(*data), [&](auto byte) { append_hex(byte, result); });
    return result;
}

template std::string to_string(const Hash<HashFunction::sha1>&);
template std::string to_string(const Hash<HashFunction::sha256>&);
template std::string to_string(const Hash<HashFunction::sha384>&);

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred
