/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef ZONE_DATA_HH_5F4A82EACB134ECC09CCDF32AB7173BB//date "+%s.%N"|md5sum|tr "[a-f]" "[A-F]"
#define ZONE_DATA_HH_5F4A82EACB134ECC09CCDF32AB7173BB

#include "src/impl/lib/absolute_fqdn.hh"
#include "src/impl/lib/dnskey_record.hh"
#include "src/impl/lib/ds_record.hh"
#include "src/impl/lib/zone_name.hh"

#include <boost/asio/ip/address.hpp>
#include <boost/optional.hpp>

#include <cstdint>
#include <map>
#include <set>
#include <string>
#include <vector>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

class ZoneData
{
public:
    struct SoaRecord
    {
        struct RData
        {
            using MName = StrongType<AbsoluteFqdn, struct ZoneDataSoaRecordMNameTag_>;
            using RName = StrongType<AbsoluteFqdn, struct ZoneDataSoaRecordRNameTag_>;
            using Serial = StrongType<std::uint32_t, struct ZoneDataSoaRecordSerialTag_>;
            using Refresh = StrongType<std::uint32_t, struct ZoneDataSoaRecordRefreshTag_>;
            using Retry = StrongType<std::uint32_t, struct ZoneDataSoaRecordRetryTag_>;
            using Expire = StrongType<std::uint32_t, struct ZoneDataSoaRecordExpireTag_>;
            using Minimum = StrongType<std::uint32_t, struct ZoneDataSoaRecordMinimumTag_>;
            MName m_name;
            RName r_name;
            Serial serial;
            Refresh refresh;
            Retry retry;
            Expire expire;
            Minimum minimum;
        };
        ZoneName zone;
        RData r_data;
    };
    struct Common
    {
        using Ttl = StrongType<std::uint32_t, struct ZoneDataCommonTtlTag_>;
        Optional<Ttl> ttl;
        SoaRecord soa_record;
        std::map<AbsoluteFqdn, std::vector<boost::asio::ip::address>> nameservers;
    };
    const Common& get_common() const noexcept;

    struct Context
    {
        explicit Context(const Common& common) noexcept;

        using KeysetId = StrongNumericId<std::int64_t, struct ZoneDataContextKeysetIdTag_>;
        using NssetId = StrongNumericId<std::int64_t, struct ZoneDataContextNssetIdTag_>;
        struct Domain
        {
            explicit Domain(AbsoluteFqdn fqdn);
            AbsoluteFqdn fqdn;
            Optional<KeysetId> keyset_id;
            NssetId nsset_id;
        };
        using DomainList = std::vector<Domain>;

        struct Keyset
        {
            KeysetId keyset_id;
            DnsKey dns_key;
        };
        using KeysetList = std::vector<Keyset>;

        struct Nsset
        {
            NssetId nsset_id;
            using Nameservers = std::map<AbsoluteFqdn, std::vector<boost::asio::ip::address>>;
            Nameservers nameservers;
        };
        using NssetList = std::vector<Nsset>;

        Common common;
        DomainList domains;
        KeysetList keysets;
        NssetList nssets;
    };

    class Nameserver
    {
    public:
        struct GlueRecords
        {
            std::vector<boost::asio::ip::address_v4> a_records;
            std::vector<boost::asio::ip::address_v6> aaaa_records;
        };
        const AbsoluteFqdn& get_nameserver_hostname() const noexcept;
        bool are_glue_records_permitted() const noexcept;
        const GlueRecords& get_glue_records() const;
    private:
        explicit Nameserver(const AbsoluteFqdn& domain, const Context::Nsset::Nameservers::value_type& nameserver_data);
        AbsoluteFqdn fqdn_;
        boost::optional<GlueRecords> glue_;
        friend class ZoneData;
    };
    class DsRData
    {
    public:
        using KeyTag = Lib::DsRecord::KeyTag;
        using Algorithm = DnsKey::Algorithm;
        using DigestType = DsRecord::DigestType;
        using Digest = DsRecord::Digest;
        KeyTag get_key_tag() const noexcept;
        Algorithm get_algorithm() const noexcept;
        DigestType get_digest_type() const noexcept;
        const Digest& get_digest() const noexcept;
    private:
        explicit DsRData(
                const AbsoluteFqdn& domain_name,
                const DnsKey& dns_key,
                DigestType digest_type);
        KeyTag key_tag_;
        Algorithm algorithm_;
        DigestType digest_type_;
        Digest digest_;
        friend class ZoneData;
    };

    struct Options
    {
        struct DsRecord
        {
            using DigestTypes = std::set<DsRData::DigestType>;
            DigestTypes digest_types;
        };
        DsRecord ds_record;
    };
    explicit ZoneData(Context ctx, Options options) noexcept;

    class ConstIterator;
    class Domain
    {
    public:
        const AbsoluteFqdn& get_domain_name() const noexcept;
        const std::vector<DsRData>& get_ds_r_data() const noexcept;
        const std::vector<Nameserver>& get_nameservers() const noexcept;
    private:
        explicit Domain(
                const Context& ctx,
                const Context::Domain& domain_data,
                const Options::DsRecord::DigestTypes& digest_types);
        explicit Domain(const AbsoluteFqdn& domain, const Context::Nsset::Nameservers& nameservers);
        AbsoluteFqdn fqdn_;
        std::vector<DsRData> ds_r_data_;
        std::vector<Nameserver> nameservers_;
        friend class ConstIterator;
        friend class ZoneData;
    };
    Domain get_zone_nameservers() const;

    class ConstIterator
    {
    public:
        ConstIterator() = delete;
        Domain operator*() const;
        ConstIterator& operator++() noexcept;
        ConstIterator operator++(int) noexcept;
    private:
        explicit ConstIterator(const ZoneData& zone_data, std::int64_t idx) noexcept;
        friend bool operator==(const ConstIterator& lhs, const ConstIterator& rhs) noexcept;
        friend bool operator!=(const ConstIterator& lhs, const ConstIterator& rhs) noexcept;
        friend bool operator<(const ConstIterator& lhs, const ConstIterator& rhs) noexcept;
        friend std::int64_t operator-(const ConstIterator& lhs, const ConstIterator& rhs) noexcept;
        const ZoneData& zone_data_;
        std::int64_t idx_;
        friend class ZoneData;
    };
    ConstIterator begin() const noexcept;
    ConstIterator end() const noexcept;
private:
    static std::vector<DsRData> make_ds_records(
            const Context::KeysetList& keyset_list,
            const Context::Domain& domain_data,
            const Options::DsRecord::DigestTypes& digest_types);
    static std::vector<Nameserver> make_nameservers(
            const Context::NssetList& nsset_list,
            const Context::Domain& domain_data);
    static std::vector<Nameserver> make_nameservers(
            const AbsoluteFqdn& domain,
            const Context::Nsset::Nameservers& nameservers);
    const Context ctx_;
    const Options options_;
};

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred

#endif//ZONE_DATA_HH_5F4A82EACB134ECC09CCDF32AB7173BB
