/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "test/client/cfg.hh"

#include <cstdlib>
#include <fstream>
#include <iostream>
#include <unordered_map>

namespace Test {
namespace Cfg {

namespace {

//critical/error/warning/info/debug/trace
decltype(auto) make_severity(const std::string& str)
{
    static const std::unordered_map<std::string, LibLog::Level> str2level =
    {
        { "critical", LibLog::Level::critical },
        { "error", LibLog::Level::error },
        { "warning", LibLog::Level::warning },
        { "info", LibLog::Level::info },
        { "debug", LibLog::Level::debug },
        { "trace", LibLog::Level::trace }
    };
    const auto level_itr = str2level.find(str);
    if (level_itr != str2level.end())
    {
        return level_itr->second;
    }
    throw Exception("\"" + str + "\" not a valid severity");
}

constexpr int fake_argc = -1;
constexpr const char* const fake_argv[] = { nullptr };

}//namespace Cfg::{anonymous}

const Options& Options::get()
{
    return init(fake_argc, fake_argv);
}

const Options& Options::init(int argc, const char* const* argv)
{
    static const Options* singleton_ptr = nullptr;
    const bool init_requested = (argc != fake_argc) || (argv != fake_argv);
    const bool first_run = singleton_ptr == nullptr;
    if (first_run)
    {
        if (!init_requested)
        {
            throw std::runtime_error("First call of Cfg::Options::init must contain valid arguments");
        }
        static const Options singleton{argc, argv};
        singleton_ptr = &singleton;
    }
    else if (init_requested)
    {
        throw std::runtime_error("Only first call of Cfg::Options::init can contain valid arguments");
    }
    return *singleton_ptr;
}

namespace {

constexpr struct OptionNames
{
    static constexpr struct ZoneService
    {
        static constexpr auto connect_to = "zone_service.connect_to";
    } zone_service = {};
    static constexpr auto zone = "zone";
} option{};

static constexpr unsigned terminal_width = 120;

decltype(auto) make_generic_options(boost::optional<std::string>& config_file_name)
{
    boost::program_options::options_description generic_options{"Generic options", terminal_width};
    const auto set_config_file_name = [&](const std::string& file_name) { config_file_name = file_name; };
    generic_options.add_options()
        ("help,h", "produce help message")
        ("config,c",
         boost::program_options::value<std::string>()->default_value(std::string{"client-zone.conf"})->notifier(set_config_file_name),
         "name of a file of a configuration.");
    return generic_options;
}

struct CommonLogOptions
{
    enum class Device
    {
        console,
        file
    };
    static decltype(auto) make_device(const std::string& device)
    {
        static const std::unordered_map<std::string, Device> str2device =
            {
                { "console", Device::console },
                { "file", Device::file }
            };
        const auto device_itr = str2device.find(device);
        if (device_itr != str2device.end())
        {
            return device_itr->second;
        }
        throw Exception{"\"" + device + "\" is not a valid device"};
    }
    std::set<Device> devices;
    boost::optional<LibLog::Level> min_severity;
    boost::optional<bool> log_grpc_library_events;
    void set_devices(const std::vector<std::string>& str_devices)
    {
        std::for_each(
            str_devices.begin(),
            str_devices.end(),
            [&](const std::string& str_device)
            {
                devices.insert(make_device(str_device));
            });
    }
    void set_min_severity(const std::string& severity)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(severity);
    }
    void set_log_grpc_library_events(bool value)
    {
        log_grpc_library_events = value;
    }
};

decltype(auto) make_log_options(CommonLogOptions& common_log_options)
{
    boost::program_options::options_description options_description{"Logging options", terminal_width};
    const auto set_devices = [&](const std::vector<std::string>& devices) { common_log_options.set_devices(devices); };
    const auto set_min_severity = [&](const std::string& severity) { common_log_options.set_min_severity(severity); };
    const auto set_log_grpc_library_events = [&](bool to_log) { common_log_options.set_log_grpc_library_events(to_log); };
    options_description.add_options()
        ("log.device", boost::program_options::value<std::vector<std::string>>()->multitoken()->notifier(set_devices), "where to log (console/file/syslog)")
        ("log.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace")
        ("log.grpc_library",
         boost::program_options::value<bool>()->zero_tokens()->notifier(set_log_grpc_library_events),
         "log gRPC library events");
    return options_description;
}

//stderr/stdout
decltype(auto) make_output_stream(const std::string& str)
{
    static const std::unordered_map<std::string, LibLog::Sink::ConsoleSinkConfig::OutputStream> str2output_stream =
        {
            { "stderr", LibLog::Sink::ConsoleSinkConfig::OutputStream::stderr },
            { "stdout", LibLog::Sink::ConsoleSinkConfig::OutputStream::stdout }
        };
    const auto output_stream_itr = str2output_stream.find(str);
    if (output_stream_itr != str2output_stream.end())
    {
        return output_stream_itr->second;
    }
    throw Exception("\"" + str + "\" not a valid output stream");
}

//always/auto/never
decltype(auto) make_color_mode(const std::string& str)
{
    static const std::unordered_map<std::string, LibLog::ColorMode> str2color_mode =
        {
            { "always", LibLog::ColorMode::always },
            { "auto", LibLog::ColorMode::automatic },
            { "never", LibLog::ColorMode::never }
        };
    const auto color_mode_itr = str2color_mode.find(str);
    if (color_mode_itr != str2color_mode.end())
    {
        return color_mode_itr->second;
    }
    throw Exception("\"" + str + "\" not a valid color mode");
}

struct ConsoleLogOptions
{
    boost::optional<LibLog::Level> min_severity;
    boost::optional<LibLog::Sink::ConsoleSinkConfig::OutputStream> output_stream;
    boost::optional<LibLog::ColorMode> color_mode;
    void set_min_severity(const std::string& value)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(value);
    }
    void set_output_stream(const std::string& value)
    {
        if (output_stream != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        output_stream = make_output_stream(value);
    }
    void set_color_mode(const std::string& value)
    {
        if (color_mode != boost::none)
        {
            throw Exception{"color mode can be defined at most once"};
        }
        color_mode = make_color_mode(value);
    }
};

decltype(auto) make_log_console_options(ConsoleLogOptions& console_log_options)
{
    boost::program_options::options_description options_description{"Logging on console options", terminal_width};
    const auto set_min_severity = [&](const std::string& severity) { console_log_options.set_min_severity(severity); };
    const auto set_output_stream = [&](const std::string& output_stream) { console_log_options.set_output_stream(output_stream); };
    const auto set_color_mode = [&](const std::string& color_mode) { console_log_options.set_color_mode(color_mode); };
    options_description.add_options()
        ("log.console.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace")
        ("log.console.output_stream",
         boost::program_options::value<std::string>()->notifier(set_output_stream),
         "where to log (stderr/stdout), default is stderr")
        ("log.console.color_mode",
         boost::program_options::value<std::string>()->notifier(set_color_mode),
         "how to colorize output (always/auto/never), default is never");
    return options_description;
}

struct FileLogOptions
{
    boost::optional<std::string> file_name;
    boost::optional<LibLog::Level> min_severity;
    void set_file_name(const std::string& name)
    {
        if (file_name != boost::none)
        {
            throw Exception{"file name can be defined at most once"};
        }
        file_name = name;
    }
    void set_min_severity(const std::string& severity)
    {
        if (min_severity != boost::none)
        {
            throw Exception{"severity can be defined at most once"};
        }
        min_severity = make_severity(severity);
    }
};

decltype(auto) make_log_file_options(FileLogOptions& file_log_options)
{
    boost::program_options::options_description options_description{"Logging into file options", terminal_width};
    const auto set_file_name = [&](const std::string& file_name) { file_log_options.set_file_name(file_name); };
    const auto set_min_severity = [&](const std::string& severity) { file_log_options.set_min_severity(severity); };
    options_description.add_options()
        ("log.file.file_name", boost::program_options::value<std::string>()->notifier(set_file_name), "what file to log into")
        ("log.file.min_severity",
         boost::program_options::value<std::string>()->notifier(set_min_severity),
         "do not log more trivial events; "
         "severity in descending order: critical/error/warning/info/debug/trace");
    return options_description;
}

decltype(auto) make_zone_service_options(Options::ZoneService& service)
{
    boost::program_options::options_description service_options{"Zone service options", terminal_width};
    service_options.add_options()
        (option.zone_service.connect_to, boost::program_options::value<std::string>(&service.connect_to), "server's listen address");
    return service_options;
}

decltype(auto) make_zone_options(Options::Zone& zone)
{
    boost::program_options::options_description zone_options{"Zone options", terminal_width};
    zone_options.add_options()
        (option.zone, boost::program_options::value<std::string>(&zone.fqdn), "zone which data are requested");
    return zone_options;
}

decltype(auto) make_zone_file_options(Options::ZoneFile& zone_file)
{
    boost::program_options::options_description zone_file_options{"Zone file options", terminal_width};
    zone_file_options.add_options()
        ("zone_file.file_name,f", boost::program_options::value<std::string>(&zone_file.file_name)->default_value("-"), "compare zone data with zone file");
    return zone_file_options;
}

void collect_log_options(
    const CommonLogOptions& common_log_options,
    const ConsoleLogOptions& console_log_options,
    const FileLogOptions& file_log_options,
    Options::Log& log_options)
{
    log_options.log_grpc_library_events = (common_log_options.log_grpc_library_events != boost::none) && *common_log_options.log_grpc_library_events;

    static constexpr auto default_log_min_severity = LibLog::Level::critical;
    const auto common_log_min_severity = common_log_options.min_severity != boost::none ? *common_log_options.min_severity
                                                                                        : default_log_min_severity;
    std::for_each(
        common_log_options.devices.begin(),
        common_log_options.devices.end(),
        [&](CommonLogOptions::Device device)
        {
            switch (device)
            {
                case CommonLogOptions::Device::console:
                {
                    Options::Log::Console console;
                    console.min_severity = console_log_options.min_severity != boost::none ? *console_log_options.min_severity
                                                                                           : common_log_min_severity;
                    static constexpr auto default_output_stream = LibLog::Sink::ConsoleSinkConfig::OutputStream::stderr;
                    console.output_stream = console_log_options.output_stream != boost::none ? *console_log_options.output_stream
                                                                                             : default_output_stream;
                    static constexpr auto default_color_mode = LibLog::ColorMode::never;
                    console.color_mode = console_log_options.color_mode != boost::none ? *console_log_options.color_mode
                                                                                       : default_color_mode;
                    log_options.console = console;
                    return;
                }
                case CommonLogOptions::Device::file:
                {
                    if (file_log_options.file_name == boost::none)
                    {
                        throw MissingOption{"missing option: 'log.file.file_name'"};
                    }
                    Options::Log::File log_file;
                    log_file.file_name = *file_log_options.file_name;
                    log_file.min_severity = file_log_options.min_severity != boost::none ? *file_log_options.min_severity
                                                                                         : common_log_min_severity;
                    log_options.file = log_file;
                    return;
                }
            }
            throw Exception("Invalid device");
        });
}

}//namespace Cfg::Test::{anonymous}

Options::Options(int argc, const char* const* argv)
{
    CommonLogOptions common_log_options;
    ConsoleLogOptions console_log_options;
    FileLogOptions file_log_options;
    const boost::program_options::options_description generic_options = make_generic_options(config_file_name);
    const boost::program_options::options_description log_options = make_log_options(common_log_options);
    const boost::program_options::options_description log_console_options = make_log_console_options(console_log_options);
    const boost::program_options::options_description log_file_options = make_log_file_options(file_log_options);
    const boost::program_options::options_description service_options = make_zone_service_options(zone_service);
    const boost::program_options::options_description zone_options = make_zone_options(zone);
    const boost::program_options::options_description zone_file_options = make_zone_file_options(zone_file);

    boost::program_options::options_description command_line_options{"test-client options", terminal_width};
    command_line_options
        .add(generic_options)
        .add(log_options)
        .add(log_console_options)
        .add(log_file_options)
        .add(service_options)
        .add(zone_options)
        .add(zone_file_options);
    boost::program_options::options_description config_file_options;
    config_file_options
        .add(log_options)
        .add(log_console_options)
        .add(log_file_options)
        .add(service_options);

    boost::program_options::variables_map variables_map{};
    boost::program_options::positional_options_description pos_options{};
    pos_options.add(option.zone, 1);
    const auto has_option = [&](const char* option) { return 0 < variables_map.count(option); };
    try
    {
        boost::program_options::store(
                boost::program_options::command_line_parser(argc, argv)
                    .options(command_line_options)
                    .positional(pos_options)
                    .run(),
                variables_map);
    }
    catch (const boost::program_options::unknown_option& unknown_option)
    {
        std::ostringstream out;
        out << unknown_option.what() << "\n\n" << command_line_options;
        throw UnknownOption(out.str());
    }
    boost::program_options::notify(variables_map);

    if (has_option("help"))
    {
        std::ostringstream out;
        out << command_line_options;
        throw HelpRequested(out.str());
    }
    const bool config_file_name_presents = config_file_name != boost::none;
    if (config_file_name_presents)
    {
        std::ifstream config_file{*config_file_name};
        if (!config_file)
        {
            throw Exception("can not open config file \"" + *config_file_name + "\"");
        }
        try
        {
            boost::program_options::store(
                    boost::program_options::parse_config_file(
                            config_file,
                            config_file_options),
                    variables_map);
        }
        catch (const boost::program_options::unknown_option& unknown_option)
        {
            std::ostringstream out;
            out << unknown_option.what() << "\n\n" << command_line_options;
            throw UnknownOption(out.str());
        }
        boost::program_options::notify(variables_map);
    }

    collect_log_options(
        common_log_options,
        console_log_options,
        file_log_options,
        log);

    const auto required = [&](const char* variable)
    {
        if (variables_map.count(variable) == 0)
        {
            throw MissingOption{"missing option: '" + std::string{variable} + "'"};
        }
    };

    required(option.zone_service.connect_to);
    required(option.zone);
}

HelpRequested::HelpRequested(std::string msg)
    : msg_{std::move(msg)}
{ }

const char* HelpRequested::what() const noexcept
{
    return msg_.c_str();
}

Exception::Exception(std::string msg)
    : std::runtime_error{std::move(msg)}
{ }

UnknownOption::UnknownOption(std::string msg)
    : Exception{std::move(msg)}
{ }

MissingOption::MissingOption(std::string msg)
    : Exception{std::move(msg)}
{ }

void handle_cli_args(int argc, char** argv)
{
    try
    {
        Cfg::Options::init(argc, argv);
    }
    catch (const Cfg::HelpRequested& help)
    {
        std::cout << help.what() << std::endl;
        ::exit(EXIT_SUCCESS);
    }
    catch (const Cfg::UnknownOption& unknown_option)
    {
        std::cerr << unknown_option.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (const Cfg::MissingOption& missing_option)
    {
        std::cerr << missing_option.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (const Cfg::Exception& e)
    {
        std::cerr << "Configuration problem: " << e.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        ::exit(EXIT_FAILURE);
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        ::exit(EXIT_FAILURE);
    }
}

}//namespace Test::Cfg
}//namespace Test
