/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "test/client/get_zone_data.hh"

#include "liblog/liblog.hh"

#include "fred_api/zone/service_zone_grpc.grpc.pb.h"

#include <grpc++/grpc++.h>

#include <algorithm>
#include <fstream>
#include <functional>
#include <iostream>
#include <memory>
#include <regex>
#include <stdexcept>
#include <string>
#include <tuple>
#include <utility>

namespace Test {

namespace {

void set_common(const Fred::Zone::Api::ZoneDataReply::Data& data, Test::ZoneData& zone_data)
{
    if (data.OptionalTtl_case() == Fred::Zone::Api::ZoneDataReply::Data::kTtl)
    {
        zone_data.ttl = data.ttl();
    }
    if (!zone_data.origin.empty())
    {
        struct TooManySoaRecords : std::runtime_error
        {
            TooManySoaRecords() : std::runtime_error{std::string{}} { }
            const char* what() const noexcept override { return "too many SOA records"; }
        };
        throw TooManySoaRecords{};
    }
    zone_data.origin = data.soa_record().zone().absolute_fqdn();
    const auto& r_data = data.soa_record().rdata();
    zone_data.soa_record = ZoneData::SoaRecord
    {
        r_data.mname().absolute_fqdn(),
        r_data.rname().absolute_fqdn(),
        r_data.serial(),
        r_data.refresh(),
        r_data.retry(),
        r_data.expire(),
        r_data.minimum()
    };
}

#if 106600 <= BOOST_VERSION
decltype(auto) make_address_v4(const std::string& addr)
{
    return boost::asio::ip::make_address_v4(addr);
}

decltype(auto) make_address_v6(const std::string& addr)
{
    return boost::asio::ip::make_address_v6(addr);
}
#else
decltype(auto) make_address_v4(const std::string& addr)
{
    return boost::asio::ip::address_v4::from_string(addr);
}

decltype(auto) make_address_v6(const std::string& addr)
{
    return boost::asio::ip::address_v6::from_string(addr);
}
#endif

void add_glue_record(const Fred::Zone::Api::DomainRecord::ARdata& a, std::set<boost::asio::ip::address_v4>& addresses)
{
    const bool a_record_is_unique = addresses.insert(make_address_v4(a.ipv4().address())).second;
    if (!a_record_is_unique)
    {
        struct MultipleARecord : std::runtime_error
        {
            MultipleARecord() : std::runtime_error{std::string{}} { }
            const char* what() const noexcept override { return "A records must be unique"; }
        };
        throw MultipleARecord{};
    }
}

void add_glue_record(const Fred::Zone::Api::DomainRecord::AaaaRdata& aaaa, std::set<boost::asio::ip::address_v6>& addresses)
{
    const bool aaaa_record_is_unique = addresses.insert(make_address_v6(aaaa.ipv6().address())).second;
    if (!aaaa_record_is_unique)
    {
        struct MultipleAaaaRecord : std::runtime_error
        {
            MultipleAaaaRecord() : std::runtime_error{std::string{}} { }
            const char* what() const noexcept override { return "AAAA records must be unique"; }
        };
        throw MultipleAaaaRecord{};
    }
}

void add_record(const Fred::Zone::Api::DomainRecord& record, Test::ZoneData& zone_data)
{
    std::for_each(record.ns_records().begin(), record.ns_records().end(), [&](const Fred::Zone::Api::DomainRecord::NsRdata& rdata)
    {
        ZoneData::DomainData& domain = zone_data.domains[record.owner().absolute_fqdn()];
        const bool unique_ns = domain.nameservers.insert(rdata.nameserver().absolute_fqdn()).second;
        if (!unique_ns)
        {
            struct MultipleNameservers : std::runtime_error
            {
                MultipleNameservers() : std::runtime_error{std::string{}} { }
                const char* what() const noexcept override { return "nameservers must be unique"; }
            };
            throw MultipleNameservers{};
        }
    });
    std::for_each(record.ds_records().begin(), record.ds_records().end(), [&](const Fred::Zone::Api::DomainRecord::DsRdata& rdata)
    {
        ZoneData::DomainData& domain = zone_data.domains[record.owner().absolute_fqdn()];
        domain.ds_records.insert(ZoneData::DomainData::DsRecord{
                static_cast<std::uint16_t>(rdata.key_tag()),
                static_cast<std::uint8_t>(rdata.algorithm()),
                static_cast<std::uint8_t>(rdata.digest_type()),
                rdata.digest().hexa_value()});
    });
    std::for_each(record.a_records().begin(), record.a_records().end(), [&](const Fred::Zone::Api::DomainRecord::ARdata& rdata)
    {
        add_glue_record(rdata, zone_data.a_records[record.owner().absolute_fqdn()]);
    });
    std::for_each(record.aaaa_records().begin(), record.aaaa_records().end(), [&](const Fred::Zone::Api::DomainRecord::AaaaRdata& rdata)
    {
        add_glue_record(rdata, zone_data.aaaa_records[record.owner().absolute_fqdn()]);
    });
}

bool read_data_from_stdin(const Cfg::Options::ZoneFile& options)
{
    return options.file_name.empty() || (options.file_name == "-");
}

class ZoneFile
{
public:
    explicit ZoneFile(const Cfg::Options::ZoneFile& options)
        : stream_{read_data_from_stdin(options) ? nullptr
                                                : std::make_unique<std::ifstream>(options.file_name)}
    { }
    std::istream& get_stream() const noexcept
    {
        if (stream_ != nullptr)
        {
            return *stream_;
        }
        return std::cin;
    }
private:
    std::unique_ptr<std::ifstream> stream_;
};

}//namespace Test::{anonymous}

ZoneData get_zone_data(const Cfg::Options::ZoneService& zone_service_options, const Cfg::Options::Zone& zone_options)
{
    auto channel = grpc::CreateChannel(zone_service_options.connect_to, grpc::InsecureChannelCredentials());
    ::grpc::ClientContext ctx;
    Fred::Zone::Api::ZoneDataRequest request;
    *(request.mutable_zone()->mutable_fqdn()) = zone_options.fqdn;
    auto reader = Fred::Zone::Api::Zone::NewStub(channel)->get_zone_data(&ctx, request);
    LIBLOG_DEBUG("have a reader");
    ZoneData data{};
    Fred::Zone::Api::ZoneDataReply reply;
    auto read_cnt = 0;
    while (reader->Read(&reply))
    {
        if (reply.data().has_soa_record())
        {
            set_common(reply.data(), data);
        }
        std::for_each(std::begin(reply.data().records()), std::end(reply.data().records()), [&](const Fred::Zone::Api::DomainRecord& record)
        {
            add_record(record, data);
        });
        if (read_cnt == 0)
        {
            LIBLOG_DEBUG("first read done");
        }
        ++read_cnt;
    }
    LIBLOG_DEBUG("last ({}.) read done", read_cnt);
    const auto status = reader->Finish();
    if (status.ok())
    {
        return data;
    }
    switch (status.error_code())
    {
        case grpc::StatusCode::NOT_FOUND:
            LIBLOG_INFO("get_zone_data failure: {}", status.error_message());
            break;
        default:
            LIBLOG_ERROR("get_zone_data failure: {}", status.error_message());
            break;
    }
    struct GetZoneDataFailed : std::runtime_error
    {
        GetZoneDataFailed() : std::runtime_error{std::string{}} { }
        const char* what() const noexcept override { return "get_zone_data failed"; }
    };
    throw GetZoneDataFailed{};
}

ZoneData get_zone_data(const Cfg::Options::ZoneFile& zone_file_options)
{
    const ZoneFile zone_file{zone_file_options};
    LIBLOG_DEBUG("file is ready");
    ZoneData zone_data{};
    std::string line{};
    auto line_number = std::int64_t{0};
    auto match_cnt = std::int64_t{0};
    static const auto make_regex = [](const char* regex)
    {
        constexpr auto regex_options = std::regex_constants::extended | std::regex_constants::optimize;
        return std::regex{regex, regex_options};
    };
    //$TTL 3600 ;default TTL for all records in zone
    const auto regex_ttl = make_regex(
        "\\$TTL[[:space:]]+([0-9]+)[[:space:]]*(;.*)?");

    //;--- domain records ---
    const auto regex_empty_line = make_regex(
        "[[:space:]]*(;.*)?");

    //gmup1z7efh0kw.cz. IN NS ns1.ipnet.cz
    const auto regex_any_in_record = make_regex(
        "([^[:space:];]+)?[[:space:]]+IN[[:space:]]+(NS|DS|A|AAAA)[[:space:]]+(([[:space:]]*[^[:space:];]+)+)[[:space:]]*(;.*)?");

    //cz. IN SOA a.ns.nic.cz. hostmaster.nic.cz. (1604034083 900 300 604800 900)
    const auto regex_soa_record = make_regex(
        "([^[:space:];]+)[[:space:]]+IN[[:space:]]+SOA[[:space:]]+"
        "([^[:space:];]+)[[:space:]]+([^[:space:];]+)[[:space:]]+"
        "\\([[:space:]]*([0-9]+)[[:space:]]+([0-9]+)[[:space:]]+([0-9]+)[[:space:]]+([0-9]+)[[:space:]]+([0-9]+)[[:space:]]*\\)"
        "[[:space:]]*(;.*)?");

    //32691 10 1 83806D6E448A02A54F6391A9EDA68688F3A48569
    const auto regex_ds_record_data = make_regex(
        "([0-9]+)[[:space:]]+([0-9]+)[[:space:]]+([0-9]+)[[:space:]]+([0-9a-fA-F]+)");

    auto last_owner = std::string{};
    const auto get_owner = [&](const std::ssub_match& match)
    {
        if (match.matched)
        {
            last_owner = match.str();
        }
        return last_owner;
    };
    while (getline(zone_file.get_stream(), line))
    {
        ++line_number;
        std::smatch match_iter{};
        if (std::regex_match(line, match_iter, regex_any_in_record))
        {
            const auto owner = get_owner(match_iter[1]);
            const auto record_type = match_iter[2].str();
            const auto record_data = match_iter[3].str();
            if (record_type == "NS")
            {
                auto& domain_data = zone_data.domains[owner];
                domain_data.nameservers.insert(record_data);
            }
            else if (record_type == "DS")
            {
                if (!std::regex_match(record_data, match_iter, regex_ds_record_data))
                {
                    struct SyntaxError : std::runtime_error
                    {
                        SyntaxError(std::int64_t line)
                            : std::runtime_error{"syntax error at line " + std::to_string(line)}
                        { }
                    };
                    throw SyntaxError{line_number};
                }
                auto& domain_data = zone_data.domains[owner];
                auto ds_record = ZoneData::DomainData::DsRecord{
                        static_cast<std::uint16_t>(std::stoul(match_iter[1].str())),
                        static_cast<std::uint8_t>(std::stoul(match_iter[2].str())),
                        static_cast<std::uint8_t>(std::stoul(match_iter[3].str())),
                        match_iter[4].str()
                };
                const bool ds_record_is_unique = domain_data.ds_records.insert(std::move(ds_record)).second;
                if (!ds_record_is_unique)
                {
                    struct MultipleDsRecord : std::runtime_error
                    {
                        MultipleDsRecord() : std::runtime_error{std::string{}} { }
                        const char* what() const noexcept override { return "DS records must be unique"; }
                    };
                    throw MultipleDsRecord{};
                }
            }
            else if (record_type == "AAAA")
            {
                auto& aaaa_records = zone_data.aaaa_records[owner];
                const bool aaaa_record_is_unique= aaaa_records.insert(make_address_v6(record_data)).second;
                if (!aaaa_record_is_unique)
                {
                    struct MultipleAaaaRecord : std::runtime_error
                    {
                        MultipleAaaaRecord() : std::runtime_error{std::string{}} { }
                        const char* what() const noexcept override { return "AAAA records must be unique"; }
                    };
                    throw MultipleAaaaRecord{};
                }
            }
            else
            {
                auto& a_records = zone_data.a_records[owner];
                const bool a_record_is_unique= a_records.insert(make_address_v4(record_data)).second;
                if (!a_record_is_unique)
                {
                    struct MultipleARecord : std::runtime_error
                    {
                        MultipleARecord() : std::runtime_error{std::string{}} { }
                        const char* what() const noexcept override { return "A records must be unique"; }
                    };
                    throw MultipleARecord{};
                }
            }
            ++match_cnt;
        }
        else if (std::regex_match(line, match_iter, regex_ttl))
        {
            zone_data.ttl = std::stoul(match_iter[1].str());
            LIBLOG_DEBUG("ttl = {}", zone_data.ttl);
            ++match_cnt;
        }
        else if (std::regex_match(line, match_iter, regex_soa_record))
        {
            zone_data.origin = match_iter[1].str();
            last_owner = zone_data.origin;
            zone_data.soa_record = ZoneData::SoaRecord{
                    match_iter[2].str(),
                    match_iter[3].str(),
                    static_cast<std::uint32_t>(std::stoul(match_iter[4].str())),
                    static_cast<std::uint32_t>(std::stoul(match_iter[5].str())),
                    static_cast<std::uint32_t>(std::stoul(match_iter[6].str())),
                    static_cast<std::uint32_t>(std::stoul(match_iter[7].str())),
                    static_cast<std::uint32_t>(std::stoul(match_iter[8].str()))
            };
            LIBLOG_DEBUG("{} SOA {} {} ({} {} {} {} {})", zone_data.origin, zone_data.soa_record.m_name, zone_data.soa_record.r_name, zone_data.soa_record.serial, zone_data.soa_record.refresh, zone_data.soa_record.retry, zone_data.soa_record.expire, zone_data.soa_record.minimum);
            ++match_cnt;
        }
        else if (!std::regex_match(line, match_iter, regex_empty_line))
        {
            struct SyntaxError : std::runtime_error
            {
                SyntaxError(std::int64_t line)
                    : std::runtime_error{"syntax error at line " + std::to_string(line)}
                { }
            };
            throw SyntaxError{line_number};
        }
    }
    LIBLOG_DEBUG("zone file has {} lines total and {} data lines", line_number, match_cnt);
    return zone_data;
}

namespace {

void log(const Diff& diff, const char* entity)
{
    if (diff.change != boost::none)
    {
        if (0 < diff.change->removed)
        {
            LIBLOG_WARNING("{} {} removed", diff.change->removed, entity);
        }
        if (0 < diff.change->added)
        {
            LIBLOG_WARNING("{} {} added", diff.change->added, entity);
        }
    }
    LIBLOG_INFO("{} {} unchanged", diff.unchanged, entity);
}

template <typename T>
using Iter = decltype(std::declval<const T&>().cbegin());

template <typename T>
using Ref = decltype(*std::declval<const T&>().cbegin());

void inc_unchanged(Diff& diff)
{
    ++diff.unchanged;
}

void inc_removed(Diff& diff, std::uint64_t cnt = 1)
{
    if (diff.change == boost::none)
    {
        diff.change = Diff::Change{cnt, 0};
        return;
    }
    diff.change->removed += cnt;;
}

void inc_added(Diff& diff, std::uint64_t cnt = 1)
{
    if (diff.change == boost::none)
    {
        diff.change = Diff::Change{0, cnt};
        return;
    }
    diff.change->added += cnt;;
}

template <typename S>
void diff(
        const S& lhs,
        const S& rhs,
        std::function<bool(Iter<S>, Iter<S>)> less,
        Diff& result,
        std::function<void(Ref<S>, Ref<S>)> on_equal = [](Ref<S>, Ref<S>){},
        std::function<void(Ref<S>)> on_lhs_absent = [](Ref<S>){},
        std::function<void(Ref<S>)> on_rhs_absent = [](Ref<S>){})
{
    auto lhs_iter = lhs.begin();
    const auto lhs_end = lhs.cend();
    auto rhs_iter = rhs.begin();
    const auto rhs_end = rhs.cend();
    while ((lhs_iter != lhs_end) && (rhs_iter != rhs_end))
    {
        if (less(lhs_iter, rhs_iter))
        {
            inc_added(result);
            on_rhs_absent(*lhs_iter);
            ++lhs_iter;
            continue;
        }
        if (less(rhs_iter, lhs_iter))
        {
            inc_removed(result);
            on_lhs_absent(*rhs_iter);
            ++rhs_iter;
            continue;
        }
        inc_unchanged(result);
        on_equal(*lhs_iter, *rhs_iter);
        ++lhs_iter;
        ++rhs_iter;
    }
    while (lhs_iter != lhs_end)
    {
        inc_added(result);
        on_rhs_absent(*lhs_iter);
        ++lhs_iter;
    }
    while (rhs_iter != rhs_end)
    {
        inc_removed(result);
        on_lhs_absent(*rhs_iter);
        ++rhs_iter;
    }
}

template <typename T>
DiffValues<T> make_diff_values(const T& lhs, const T& rhs)
{
    if (lhs == rhs)
    {
        return DiffValues<T>{boost::none};
    }
    return DiffValues<T>{std::make_pair(lhs, rhs)};
}

}//namespace Test::{anonymous}

Diff::Diff() noexcept
    : unchanged{0},
      change{boost::none}
{ }

void ZoneDiff::log() const
{
    if (origin != boost::none)
    {
        LIBLOG_WARNING("file zone origin {} differs from grpc zone origin {}", origin->second, origin->first);
    }
    else
    {
        LIBLOG_INFO("ORIGIN unchanged");
    }
    if (ttl != boost::none)
    {
        LIBLOG_WARNING("file zone TTL {} differs from grpc zone TTL {}", ttl->second, ttl->first);
    }
    else
    {
        LIBLOG_INFO("TTL unchanged");
    }
    if (soa_record.serial != boost::none)
    {
        if (soa_record.serial->first < soa_record.serial->second)
        {
            LIBLOG_INFO("grpc zone is {} seconds older then file zone", soa_record.serial->second - soa_record.serial->first);
        }
        else if (soa_record.serial->second < soa_record.serial->first)
        {
            LIBLOG_INFO("file zone is {} seconds older then grpc zone", soa_record.serial->first - soa_record.serial->second);
        }
    }
    else
    {
        LIBLOG_INFO("file zone and grpc zone have the same SOA serial number");
    }
    if (soa_record.m_name != boost::none)
    {
        LIBLOG_WARNING("file zone SOA MName {} differs from grpc zone SOA MName {}", soa_record.m_name->second, soa_record.m_name->first);
    }
    if (soa_record.r_name != boost::none)
    {
        LIBLOG_WARNING("file zone SOA RName {} differs from grpc zone SOA RName {}", soa_record.r_name->second, soa_record.r_name->first);
    }
    if (soa_record.refresh != boost::none)
    {
        LIBLOG_WARNING("file zone SOA refresh {} differs from grpc zone SOA refresh {}", soa_record.refresh->second, soa_record.refresh->first);
    }
    if (soa_record.retry != boost::none)
    {
        LIBLOG_WARNING("file zone SOA retry {} differs from grpc zone SOA retry {}", soa_record.retry->second, soa_record.retry->first);
    }
    if (soa_record.expire != boost::none)
    {
        LIBLOG_WARNING("file zone SOA expire {} differs from grpc zone SOA expire {}", soa_record.expire->second, soa_record.expire->first);
    }
    if (soa_record.minimum != boost::none)
    {
        LIBLOG_WARNING("file zone SOA minimum {} differs from grpc zone SOA minimum {}", soa_record.minimum->second, soa_record.minimum->first);
    }
    Test::log(domains, "domains");
    Test::log(nameservers, "nameservers");
    Test::log(ds_records, "DS records");
    Test::log(a_records, "A records");
    Test::log(aaaa_records, "AAAA records");
    LIBLOG_INFO("change ratio = {:0.6f}%", this->change_ratio() * 100);
}

double ZoneDiff::change_ratio() const noexcept
{
    const auto get_new_sum = [](const Diff& d)
    {
        return d.change == boost::none ? d.unchanged
                                       : d.unchanged + d.change->added;
    };
    const auto get_old_sum = [](const Diff& d)
    {
        return d.change == boost::none ? d.unchanged
                                       : d.unchanged + d.change->removed;
    };
    const auto new_sum = get_new_sum(nameservers) + get_new_sum(ds_records) + get_new_sum(a_records) + get_new_sum(aaaa_records);
    const auto old_sum = get_old_sum(nameservers) + get_old_sum(ds_records) + get_old_sum(a_records) + get_old_sum(aaaa_records);
    const auto unchanged_sum = nameservers.unchanged + ds_records.unchanged + a_records.unchanged + aaaa_records.unchanged;
    if ((new_sum == 0) && (old_sum == 0))
    {
        return 0.0;
    }
    if ((unchanged_sum == 0) || (new_sum == 0) || (old_sum == 0))
    {
        return 1.0;
    }
    const auto ratio_by_old = double(old_sum - unchanged_sum) / double(old_sum);
    const auto ratio_by_new = double(new_sum - unchanged_sum) / double(new_sum);
    return std::max(ratio_by_old, ratio_by_new);
}

ZoneDiff diff(const ZoneData& grpc, const ZoneData& file)
{
    ZoneDiff zone_diff;
    zone_diff.origin = make_diff_values(grpc.origin, file.origin);
    zone_diff.ttl = make_diff_values(grpc.ttl, file.ttl);
    zone_diff.soa_record.serial = make_diff_values(grpc.soa_record.serial, file.soa_record.serial);
    zone_diff.soa_record.m_name = make_diff_values(grpc.soa_record.m_name, file.soa_record.m_name);
    zone_diff.soa_record.r_name = make_diff_values(grpc.soa_record.r_name, file.soa_record.r_name);
    zone_diff.soa_record.refresh = make_diff_values(grpc.soa_record.refresh, file.soa_record.refresh);
    zone_diff.soa_record.retry = make_diff_values(grpc.soa_record.retry, file.soa_record.retry);
    zone_diff.soa_record.expire = make_diff_values(grpc.soa_record.expire, file.soa_record.expire);
    zone_diff.soa_record.minimum = make_diff_values(grpc.soa_record.minimum, file.soa_record.minimum);
    {
        diff(grpc.domains,
             file.domains,
             [](const auto& lhs_iter, const auto& rhs_iter) { return lhs_iter->first < rhs_iter->first; },
             zone_diff.domains,
             [&](const auto& lhs, const auto& rhs)
             {
                 diff(lhs.second.nameservers,
                      rhs.second.nameservers,
                      [](const auto& lhs_iter, const auto& rhs_iter) { return *lhs_iter < *rhs_iter; },
                      zone_diff.nameservers);

                 diff(lhs.second.ds_records,
                      rhs.second.ds_records,
                      [](const auto& lhs_iter, const auto& rhs_iter) { return *lhs_iter < *rhs_iter; },
                      zone_diff.ds_records);
             },
             [&](const auto& rhs)
             {
                 inc_removed(zone_diff.nameservers, rhs.second.nameservers.size());
                 inc_removed(zone_diff.ds_records, rhs.second.ds_records.size());
             },
             [&](const auto& lhs)
             {
                 inc_added(zone_diff.nameservers, lhs.second.nameservers.size());
                 inc_added(zone_diff.ds_records, lhs.second.ds_records.size());
             });
    }
    {
        Diff a_glue_records{};
        diff(grpc.a_records,
             file.a_records,
             [](const auto& lhs_iter, const auto& rhs_iter) { return lhs_iter->first < rhs_iter->first; },
             a_glue_records,
             [&](const auto& lhs, const auto& rhs)
             {
                 diff(lhs.second,
                      rhs.second,
                      [](const auto& lhs_iter, const auto& rhs_iter) { return *lhs_iter < *rhs_iter; },
                      zone_diff.a_records);
             },
             [&](const auto& rhs) { inc_removed(zone_diff.a_records, rhs.second.size()); },
             [&](const auto& lhs) { inc_added(zone_diff.a_records, lhs.second.size()); });
    }
    {
        Diff aaaa_glue_records{};
        diff(grpc.aaaa_records,
             file.aaaa_records,
             [](const auto& lhs_iter, const auto& rhs_iter) { return lhs_iter->first < rhs_iter->first; },
             aaaa_glue_records,
             [&](const auto& lhs, const auto& rhs)
             {
                 diff(lhs.second,
                      rhs.second,
                      [](const auto& lhs_iter, const auto& rhs_iter) { return *lhs_iter < *rhs_iter; },
                      zone_diff.aaaa_records);
             },
             [&](const auto& rhs) { inc_removed(zone_diff.aaaa_records, rhs.second.size()); },
             [&](const auto& lhs) { inc_added(zone_diff.aaaa_records, lhs.second.size()); });
    }
    return zone_diff;
}

bool operator<(const ZoneData::DomainData::DsRecord& lhs, const ZoneData::DomainData::DsRecord& rhs) noexcept
{
    return std::make_tuple(lhs.digest_type, lhs.key_tag, lhs.algorithm, lhs.digest) < std::make_tuple(rhs.digest_type, rhs.key_tag, rhs.algorithm, rhs.digest);
}

}//namespace Test
