/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "test/client/cfg.hh"
#include "test/client/get_zone_data.hh"

#include "liblog/liblog.hh"
#include "liblog/log.hh"

#include "fred_api/zone/service_zone_grpc.grpc.pb.h"

#include <boost/asio/ip/address.hpp>
#include <boost/optional.hpp>
#include <boost/program_options.hpp>

#include <grpc++/grpc++.h>
#include <grpc/support/log.h>

#include <algorithm>
#include <cstdint>
#include <fstream>
#include <future>
#include <stdexcept>
#include <string>
#include <thread>
#include <tuple>
#include <utility>

namespace Test {

namespace {

void set_logging(const Cfg::Options::Log& log_options);

template <typename T = void>
struct DataFromThread
{
    DataFromThread()
        : promise{},
          future{promise.get_future()}
    { }
    std::promise<T> promise;
    std::future<T> future;
};

void run_get_zone_data(const Cfg::Options::ZoneService& zone_service_options, const Cfg::Options::Zone& zone_options, std::promise<ZoneData> result)
{
    try
    {
        result.set_value(get_zone_data(zone_service_options, zone_options));
    }
    catch (...)
    {
        result.set_exception(std::current_exception());
    }
}

}//namespace Test::{anonymous}

}//namespace Test

int main(int argc, char* argv[])
{
    Test::Cfg::handle_cli_args(argc, argv);
    try
    {
        Test::set_logging(Test::Cfg::Options::get().log);
        LIBLOG_SET_CONTEXT(Ctx, log_ctx, __func__);
        Test::DataFromThread<Test::ZoneData> zone_from_thread;
        auto thread = std::thread{
                Test::run_get_zone_data,
                std::cref(Test::Cfg::Options::get().zone_service),
                std::cref(Test::Cfg::Options::get().zone),
                std::move(zone_from_thread.promise)};
        bool success = false;
        try
        {
            const auto file_zone_data = Test::get_zone_data(Test::Cfg::Options::get().zone_file);
            const auto grpc_zone_data = zone_from_thread.future.get();
            diff(grpc_zone_data, file_zone_data).log();
            success = true;
        }
        catch (const std::exception& e)
        {
            LIBLOG_WARNING("exception caught: {}", e.what());
        }
        catch (...)
        {
            LIBLOG_WARNING("unknown exception caught");
        }
        thread.join();
        return success ? EXIT_SUCCESS
                       : EXIT_FAILURE;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }
}

namespace Test {

namespace {

void disable_grpc_library_logging()
{
    static const auto dummy_log_func = [](::gpr_log_func_args*) { };
    ::gpr_set_log_function(dummy_log_func);
    static constexpr auto do_not_log = static_cast<::gpr_log_severity>(GPR_LOG_VERBOSITY_UNSET);
    ::gpr_set_log_verbosity(do_not_log);
}

void grpc_log(::gpr_log_func_args* log_args)
{
    try
    {
        switch (log_args->severity)
        {
            case GPR_LOG_SEVERITY_DEBUG:
                LIBLOG_DEBUG("{} at {}:{}", log_args->message, log_args->file, log_args->line);
                return;
            case GPR_LOG_SEVERITY_INFO:
                LIBLOG_INFO("{} at {}:{}", log_args->message, log_args->file, log_args->line);
                return;
            case GPR_LOG_SEVERITY_ERROR:
                LIBLOG_ERROR("{} at {}:{}", log_args->message, log_args->file, log_args->line);
                return;
        }
        LIBLOG_WARNING("{} at {}:{}", log_args->message, log_args->file, log_args->line);
    }
    catch (...)
    { }
}

constexpr auto no_log_grpc_severity = static_cast<::gpr_log_severity>(GPR_LOG_VERBOSITY_UNSET);

constexpr ::gpr_log_severity make_grpc_log_severity(LibLog::Level min_severity)
{
    switch (min_severity)
    {
        case LibLog::Level::critical:
            return no_log_grpc_severity;
        case LibLog::Level::error:
            return GPR_LOG_SEVERITY_ERROR;
        case LibLog::Level::warning:
        case LibLog::Level::info:
            return GPR_LOG_SEVERITY_INFO;
        case LibLog::Level::debug:
        case LibLog::Level::trace:
            return GPR_LOG_SEVERITY_DEBUG;
    }
    return no_log_grpc_severity;
}

decltype(auto) get_min_severity(const Cfg::Options::Log& log_options)
{
    boost::optional<LibLog::Level> min_severity;
    const auto update_min_severity = [&](LibLog::Level severity)
    {
        if ((min_severity == boost::none) ||
            (severity < *min_severity))
        {
            min_severity = severity;
        }
    };
    if (log_options.console != boost::none)
    {
        update_min_severity(log_options.console->min_severity);
    }
    if (log_options.file != boost::none)
    {
        update_min_severity(log_options.file->min_severity);
    }
    return min_severity;
}

void set_grpc_library_logging(const Cfg::Options::Log& log_options)
{
    if (!log_options.log_grpc_library_events)
    {
        LIBLOG_INFO("do not log gRPC library");
        disable_grpc_library_logging();
        return;
    }
    const auto min_severity = get_min_severity(log_options);
    if (min_severity == boost::none)
    {
        disable_grpc_library_logging();
        return;
    }
    ::gpr_set_log_function(grpc_log);
    ::gpr_set_log_verbosity(make_grpc_log_severity(*min_severity));
    LIBLOG_INFO("log gRPC library");
}

decltype(auto) make_sink_config(const Cfg::Options::Log::Console& log_options)
{
    LibLog::Sink::ConsoleSinkConfig sink_config;
    sink_config.set_level(log_options.min_severity);
    sink_config.set_output_stream(log_options.output_stream);
    sink_config.set_color_mode(log_options.color_mode);
    return sink_config;
}

decltype(auto) make_sink_config(const Cfg::Options::Log::File& log_options)
{
    LibLog::Sink::FileSinkConfig sink_config{log_options.file_name};
    sink_config.set_level(log_options.min_severity);
    return sink_config;
}

void set_logging(const Cfg::Options::Log& log_options)
{
    bool has_device = false;
    LibLog::LogConfig log_config;
    if (log_options.console != boost::none)
    {
        try
        {
            log_config.add_sink_config(make_sink_config(*log_options.console));
            has_device = true;
        }
        catch (...) { }
    }
    if (log_options.file != boost::none)
    {
        try
        {
            log_config.add_sink_config(make_sink_config(*log_options.file));
            has_device = true;
        }
        catch (...) { }
    }
    if (has_device)
    {
        try
        {
            LibLog::Log::start<LibLog::ThreadMode::multi_threaded>(log_config);
            set_grpc_library_logging(log_options);
            return;
        }
        catch (const std::exception& e)
        {
            std::cerr << "Setting up logging failure: " << e.what() << std::endl;
            ::exit(EXIT_FAILURE);
        }
        catch (...)
        {
            std::cerr << "Setting up logging failure: Unexpected exception caught" << std::endl;
            ::exit(EXIT_FAILURE);
        }
    }
    try
    {
        disable_grpc_library_logging();
    }
    catch (...)
    {
        std::cerr << "unable to disable gRPC library logging" << std::endl;
    }
    std::cout << "no logging is configured!?" << std::endl;
}

}//namespace Test::{anonymous}

}//namespace Test
