/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE TestZone

#include "setup.hh"

#include "src/cfg.hh"

#include "liblog/liblog.hh"
#include "liblog/log_config.hh"
#include "liblog/log.hh"

#include <boost/test/unit_test.hpp>

class GlobalFixture
{
public:
    GlobalFixture()
    {
        const auto& master_test_suite = boost::unit_test::framework::master_test_suite();
        Cfg::handle_cli_args(master_test_suite.argc, master_test_suite.argv);
        set_logging(Cfg::Options::get().log);
        LIBLOG_INFO("Test logging is on");
    }
    ~GlobalFixture()
    {
        LIBLOG_INFO("Test is done");
    }
private:
    static decltype(auto) make_sink_config(const Cfg::Options::Log::Console& log_options)
    {
        LibLog::Sink::ConsoleSinkConfig sink_config;
        sink_config.set_level(log_options.min_severity);
        sink_config.set_output_stream(log_options.output_stream);
        sink_config.set_color_mode(log_options.color_mode);
        return sink_config;
    }
    static decltype(auto) make_sink_config(const Cfg::Options::Log::File& log_options)
    {
        LibLog::Sink::FileSinkConfig sink_config{log_options.file_name};
        sink_config.set_level(log_options.min_severity);
        return sink_config;
    }
    static decltype(auto) make_sink_config(const Cfg::Options::Log::Syslog& log_options)
    {
        LibLog::Sink::SyslogSinkConfig sink_config;
        sink_config.set_syslog_facility(log_options.facility);
        sink_config.set_level(log_options.min_severity);
        sink_config.set_syslog_ident(log_options.ident);
        sink_config.set_syslog_options(log_options.options);
        return sink_config;
    }
    static void set_logging(const Cfg::Options::Log& log_options)
    {
        bool has_device = false;
        LibLog::LogConfig log_config;
        if (log_options.console != boost::none)
        {
            try
            {
                log_config.add_sink_config(make_sink_config(*log_options.console));
                has_device = true;
            }
            catch (...) { }
        }
        if (log_options.file != boost::none)
        {
            try
            {
                log_config.add_sink_config(make_sink_config(*log_options.file));
                has_device = true;
            }
            catch (...) { }
        }
        if (log_options.syslog != boost::none)
        {
            try
            {
                log_config.add_sink_config(make_sink_config(*log_options.syslog));
                has_device = true;
            }
            catch (...) { }
        }
        if (has_device)
        {
            try
            {
                LibLog::Log::start<LibLog::ThreadMode::multi_threaded>(log_config);
                return;
            }
            catch (const std::exception& e)
            {
                std::cerr << "Setting up logging failure: " << e.what() << std::endl;
                ::exit(EXIT_FAILURE);
            }
            catch (...)
            {
                std::cerr << "Setting up logging failure: Unexpected exception caught" << std::endl;
                ::exit(EXIT_FAILURE);
            }
        }
    }
};

BOOST_GLOBAL_FIXTURE(GlobalFixture);
