/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/absolute_fqdn.hh"
#include "src/impl/lib/wire_format.hh"

#include "liblog/liblog.hh"

#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE(TestAbsoluteFqdn)

BOOST_AUTO_TEST_CASE(relative_fqdn)
{
    const std::string domain = "abc.de";
    const Fred::Zone::Impl::Lib::AbsoluteFqdn absolute_domain{Fred::Zone::Impl::Lib::Fqdn{domain}};
    BOOST_CHECK_EQUAL(*absolute_domain.get_fqdn(), domain + ".");
    Fred::Zone::Impl::Lib::WireFormat wire;
    wire.append(absolute_domain);
    BOOST_CHECK_EQUAL(wire.data().size(), absolute_domain.get_fqdn()->length() + 1);
    BOOST_CHECK_EQUAL(wire.data()[0], 3);
    BOOST_CHECK_EQUAL(wire.data()[1], 'a');
    BOOST_CHECK_EQUAL(wire.data()[2], 'b');
    BOOST_CHECK_EQUAL(wire.data()[3], 'c');
    BOOST_CHECK_EQUAL(wire.data()[4], 2);
    BOOST_CHECK_EQUAL(wire.data()[5], 'd');
    BOOST_CHECK_EQUAL(wire.data()[6], 'e');
    BOOST_CHECK_EQUAL(wire.data()[7], 0);
}

BOOST_AUTO_TEST_CASE(absolute_fqdn)
{
    const std::string domain = "abc.de.";
    const Fred::Zone::Impl::Lib::AbsoluteFqdn absolute_domain{Fred::Zone::Impl::Lib::Fqdn{domain}};
    BOOST_CHECK_EQUAL(*absolute_domain.get_fqdn(), domain);
    Fred::Zone::Impl::Lib::WireFormat wire;
    wire.append(absolute_domain);
    BOOST_CHECK_EQUAL(wire.data().size(), absolute_domain.get_fqdn()->length() + 1);
    BOOST_CHECK_EQUAL(wire.data()[0], 3);
    BOOST_CHECK_EQUAL(wire.data()[1], 'a');
    BOOST_CHECK_EQUAL(wire.data()[2], 'b');
    BOOST_CHECK_EQUAL(wire.data()[3], 'c');
    BOOST_CHECK_EQUAL(wire.data()[4], 2);
    BOOST_CHECK_EQUAL(wire.data()[5], 'd');
    BOOST_CHECK_EQUAL(wire.data()[6], 'e');
    BOOST_CHECK_EQUAL(wire.data()[7], 0);
}

namespace {

bool log_exception(const std::exception& e)
{
    LIBLOG_INFO("std::exception caught: {}", e.what());
    return true;
}

auto make_absolute_fqdn(const std::string& fqdn)
{
    return Fred::Zone::Impl::Lib::AbsoluteFqdn{Fred::Zone::Impl::Lib::Fqdn{fqdn}};
}

void wire_absolute_fqdn(const Fred::Zone::Impl::Lib::AbsoluteFqdn& fqdn)
{
    Fred::Zone::Impl::Lib::WireFormat{}.append(fqdn);
}

}//namespace {anonymous}

BOOST_AUTO_TEST_CASE(fqdn_too_long)
{
    const auto label63 = std::string(63, 'a') + ".";
    const auto label58 = std::string(58, 'a') + ".";
    const auto max_domain = (label63 + label63 + label63 + label58).append("com");
    const auto oversized_domain = (label63 + label63 + label63 + "a" + label58).append("com");
    BOOST_CHECK_EQUAL(max_domain.length(), 254);
    BOOST_CHECK_EQUAL(oversized_domain.length(), 255);
    BOOST_CHECK_NO_THROW(make_absolute_fqdn(max_domain));
    BOOST_CHECK_NO_THROW(make_absolute_fqdn(max_domain + "."));
    BOOST_CHECK_EXCEPTION(make_absolute_fqdn(oversized_domain), std::exception, log_exception);
    BOOST_CHECK_EXCEPTION(make_absolute_fqdn(oversized_domain + "."), std::exception, log_exception);
}

BOOST_AUTO_TEST_CASE(label_too_long)
{
    const auto max_label = std::string(63, 'a') + ".";
    const auto oversized_label = "a" + max_label;
    auto fqdn = make_absolute_fqdn(max_label);
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));
    fqdn = make_absolute_fqdn(max_label + "com");
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));
    fqdn = make_absolute_fqdn(max_label + max_label);
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));
    fqdn = make_absolute_fqdn(max_label + max_label + "com");
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));

    fqdn = make_absolute_fqdn(oversized_label);
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(oversized_label + "com");
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(max_label + oversized_label);
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(max_label + oversized_label + "com");
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(oversized_label + max_label);
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(oversized_label + max_label + "com");
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
}

BOOST_AUTO_TEST_CASE(empty_label_not_allowed)
{
    const auto label = std::string{"abc."};
    const auto empty_label = std::string{"."};
    auto fqdn = make_absolute_fqdn(empty_label);
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));
    fqdn = make_absolute_fqdn(label);
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));
    fqdn = make_absolute_fqdn(label + label);
    BOOST_CHECK_NO_THROW(wire_absolute_fqdn(fqdn));

    fqdn = make_absolute_fqdn(empty_label + label);
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(label + empty_label);
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
    fqdn = make_absolute_fqdn(label + empty_label + label);
    BOOST_CHECK_EXCEPTION(wire_absolute_fqdn(fqdn), std::exception, log_exception);
}

BOOST_AUTO_TEST_CASE(is_subdomain_of)
{
    BOOST_CHECK(make_absolute_fqdn("sub.domain.cz").is_subdomain_of(make_absolute_fqdn("domain.cz")));
    BOOST_CHECK(make_absolute_fqdn("sub.domain.cz.").is_subdomain_of(make_absolute_fqdn("domain.cz")));
    BOOST_CHECK(make_absolute_fqdn("sub.domain.cz").is_subdomain_of(make_absolute_fqdn("domain.cz.")));
    BOOST_CHECK(make_absolute_fqdn("sub.domain.cz.").is_subdomain_of(make_absolute_fqdn("domain.cz.")));
    BOOST_CHECK(make_absolute_fqdn("sub.domain.cz.").is_subdomain_of(make_absolute_fqdn("")));
    BOOST_CHECK(make_absolute_fqdn("sub.domain.cz.").is_subdomain_of(make_absolute_fqdn(".")));

    BOOST_CHECK(!make_absolute_fqdn("domain.cz").is_subdomain_of(make_absolute_fqdn("domain.cz")));
    BOOST_CHECK(!make_absolute_fqdn(".domain.cz").is_subdomain_of(make_absolute_fqdn("domain.cz")));
    BOOST_CHECK(!make_absolute_fqdn("domain.cz").is_subdomain_of(make_absolute_fqdn("sub.domain.cz")));
    BOOST_CHECK(!make_absolute_fqdn("").is_subdomain_of(make_absolute_fqdn("domain.cz")));
    BOOST_CHECK(!make_absolute_fqdn("").is_subdomain_of(make_absolute_fqdn("")));
    BOOST_CHECK(!make_absolute_fqdn(".").is_subdomain_of(make_absolute_fqdn("")));
    BOOST_CHECK(!make_absolute_fqdn("").is_subdomain_of(make_absolute_fqdn(".")));
    BOOST_CHECK(!make_absolute_fqdn(".").is_subdomain_of(make_absolute_fqdn(".")));
}

BOOST_AUTO_TEST_SUITE_END()//TestAbsoluteFqdn
