/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "setup.hh"

#include "src/impl/lib/base64.hh"

#include "liblog/liblog.hh"

#include <boost/test/unit_test.hpp>

#include <algorithm>
#include <array>

namespace {

template <char c0, char c1, char c2, char c3> struct Base64Quartet;
template <std::uint8_t b0, std::uint8_t b1, std::uint8_t b2> struct BinTrio;
template <std::uint8_t b0, std::uint8_t b1> struct BinDuo;
template <std::uint8_t b0> struct BinMono;

template <typename, typename> struct Base64DecodeCheck;

template <char c0, char c1, char c2, char c3,
          std::uint8_t b0, std::uint8_t b1, std::uint8_t b2>
struct Base64DecodeCheck<Base64Quartet<c0, c1, c2, c3>, BinTrio<b0, b1, b2>>
{
    void operator()() const
    {
        constexpr auto base64_encoded_data = std::array<char, 4>{ c0, c1, c2, c3 };
        std::vector<std::uint8_t> data;
        data.reserve(3);
        Fred::Zone::Impl::Lib::decode_base64(base64_encoded_data.begin(), base64_encoded_data.end(), back_inserter(data));
        BOOST_REQUIRE_EQUAL(data.size(), 3);
        BOOST_CHECK_EQUAL(data[0], b0);
        BOOST_CHECK_EQUAL(data[1], b1);
        BOOST_CHECK_EQUAL(data[2], b2);
    }
};

template <char c0, char c1, char c2,
          std::uint8_t b0, std::uint8_t b1>
struct Base64DecodeCheck<Base64Quartet<c0, c1, c2, '='>, BinDuo<b0, b1>>
{
    void operator()() const
    {
        constexpr auto base64_encoded_data = std::array<char, 4>{ c0, c1, c2, '=' };
        std::vector<std::uint8_t> data;
        data.reserve(2);
        Fred::Zone::Impl::Lib::decode_base64(base64_encoded_data.begin(), base64_encoded_data.end(), back_inserter(data));
        BOOST_REQUIRE_EQUAL(data.size(), 2);
        BOOST_CHECK_EQUAL(data[0], b0);
        BOOST_CHECK_EQUAL(data[1], b1);
    }
};

template <char c0, char c1,
          std::uint8_t b0>
struct Base64DecodeCheck<Base64Quartet<c0, c1, '=', '='>, BinMono<b0>>
{
    void operator()() const
    {
        constexpr auto base64_encoded_data = std::array<char, 4>{ c0, c1, '=', '=' };
        std::vector<std::uint8_t> data;
        data.reserve(1);
        Fred::Zone::Impl::Lib::decode_base64(base64_encoded_data.begin(), base64_encoded_data.end(), back_inserter(data));
        BOOST_REQUIRE_EQUAL(data.size(), 1);
        BOOST_CHECK_EQUAL(data[0], b0);
    }
};

template <std::size_t binary_data_length>
void check(const std::string& base64_encoded_data, const std::array<std::uint8_t, binary_data_length>& corresponding_data)
{
    std::vector<std::uint8_t> data;
    data.reserve(corresponding_data.size());
    BOOST_CHECK_NO_THROW((Fred::Zone::Impl::Lib::decode_base64(base64_encoded_data.begin(), base64_encoded_data.end(), back_inserter(data))));
    BOOST_REQUIRE_EQUAL(data.size(), corresponding_data.size());
    auto corresponding_data_iter = std::begin(corresponding_data);
    std::for_each(
            std::begin(data),
            std::end(data),
            [&](auto bin)
            {
                BOOST_CHECK_EQUAL(bin, *corresponding_data_iter);
                ++corresponding_data_iter;
            });
}

}//namespace {anonymous}

BOOST_AUTO_TEST_SUITE(TestBase64)

BOOST_AUTO_TEST_CASE(decode_ok)
{
    BOOST_CHECK_NO_THROW((Base64DecodeCheck<Base64Quartet<'A', 'B', 'C', 'D'>, BinTrio<0x00, 0x10, 0x83>>{}()));
    BOOST_CHECK_NO_THROW((Base64DecodeCheck<Base64Quartet<'h', 'S', 'M', 'b'>, BinTrio<0x85, 0x23, 0x1b>>{}()));

    BOOST_CHECK_NO_THROW((Base64DecodeCheck<Base64Quartet<'A', 'B', 'A', '='>, BinDuo<0x00, 0x10>>{}()));
    BOOST_CHECK_NO_THROW((Base64DecodeCheck<Base64Quartet<'h', 'S', 'M', '='>, BinDuo<0x85, 0x23>>{}()));

    BOOST_CHECK_NO_THROW((Base64DecodeCheck<Base64Quartet<'A', 'A', '=', '='>, BinMono<0x00>>{}()));
    BOOST_CHECK_NO_THROW((Base64DecodeCheck<Base64Quartet<'h', 'Q', '=', '='>, BinMono<0x85>>{}()));

    check("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",
          std::array<std::uint8_t, 48>{0x00, 0x10, 0x83, 0x10, 0x51, 0x87, 0x20, 0x92, 0x8b, 0x30, 0xd3, 0x8f, 0x41, 0x14, 0x93, 0x51,
                                       0x55, 0x97, 0x61, 0x96, 0x9b, 0x71, 0xd7, 0x9f, 0x82, 0x18, 0xa3, 0x92, 0x59, 0xa7, 0xa2, 0x9a,
                                       0xab, 0xb2, 0xdb, 0xaf, 0xc3, 0x1c, 0xb3, 0xd3, 0x5d, 0xb7, 0xe3, 0x9e, 0xbb, 0xf3, 0xdf, 0xbf});
}

BOOST_AUTO_TEST_CASE(decode_fail)
{
    static const auto decode = [](const std::string& base64_encoded_data)
    {
        try
        {
            Fred::Zone::Impl::Lib::decode_base64(base64_encoded_data);
            LIBLOG_WARNING("no exception caught");
        }
        catch (const Fred::Zone::Impl::Lib::Base64DecoderFailure& e)
        {
            LIBLOG_INFO("Base64DecoderFailure caught: {}", e.what());
            throw;
        }
        catch (const std::exception& e)
        {
            LIBLOG_WARNING("std::exception caught: {}", e.what());
            throw;
        }
        catch (...)
        {
            LIBLOG_ERROR("unknown exception caught");
            throw;
        }
    };
    BOOST_CHECK_EXCEPTION(decode("A"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("AB"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("ABC"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("ABC="), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("AB=="), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("AB=C"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode(" BCD"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("A CD"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("AB D"), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
    BOOST_CHECK_EXCEPTION(decode("ABC "), Fred::Zone::Impl::Lib::Base64DecoderFailure, [](const auto&) { return true; });
}

BOOST_AUTO_TEST_SUITE_END()//TestBase64
