/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/absolute_fqdn.hh"
#include "src/impl/lib/dnskey_record.hh"
#include "src/impl/lib/hash.hh"
#include "src/impl/lib/wire_format.hh"

#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE(TestDnsKey)

// iana.org. 3599 IN DNSKEY 257 3 8 AwEAAb0i1jarxjgmjsOAO7IbXvhwe4yBSGWpUnjj9SLZOguEbOY+6hrs
//                                  LL5Xa9LKHDjgF/ovovtOj4t9JrKXQZ6c/kJdtaIkGcWzhACoNkum20mn
//                                  2Ja3Zk1xJgEkMpOED5I3N6IHNCALSNw8rxu3c3xVDpD6Tf10+Bcpr4sj
//                                  ge32lNeVDjVPs7D7yJSQrr0tBui4rtsMkeQ8jNT9Hna7C/5MRAxqjpue
//                                  X0Z7kEqv4vevv1Y7xDi7yloRzY/tx8BdYEVRNOQ9pbU2ERuFhUWmwT1D
//                                  ROUBwpK6husdOGbnPyGzCiHt14An5wYpj6B4Jao9DuNTgTl6xg20oyR4
//                                  6pRM1roTeIk=
// iana.org. 21599 IN DS 39817  8 2 5FB72AE84C35A35DDF7FA24EA791CD9763C4360979C0CBB6FF2C407A
//                                  F9AE5802

namespace {

const auto iana_org_public_key = Fred::Zone::Impl::Lib::Base64EncodedData{
        "AwEAAb0i1jarxjgmjsOAO7IbXvhwe4yBSGWpUnjj9SLZOguEbOY+6hrs"
        "LL5Xa9LKHDjgF/ovovtOj4t9JrKXQZ6c/kJdtaIkGcWzhACoNkum20mn"
        "2Ja3Zk1xJgEkMpOED5I3N6IHNCALSNw8rxu3c3xVDpD6Tf10+Bcpr4sj"
        "ge32lNeVDjVPs7D7yJSQrr0tBui4rtsMkeQ8jNT9Hna7C/5MRAxqjpue"
        "X0Z7kEqv4vevv1Y7xDi7yloRzY/tx8BdYEVRNOQ9pbU2ERuFhUWmwT1D"
        "ROUBwpK6husdOGbnPyGzCiHt14An5wYpj6B4Jao9DuNTgTl6xg20oyR4"
        "6pRM1roTeIk="};

auto make_absolute_fqdn(const std::string& fqdn)
{
    return Fred::Zone::Impl::Lib::AbsoluteFqdn{Fred::Zone::Impl::Lib::Fqdn{fqdn}};
}

auto make_dnskey(std::uint16_t flags, std::uint8_t protocol, std::uint8_t algorithm, const Fred::Zone::Impl::Lib::Base64EncodedData& public_key)
{
    using namespace Fred::Zone::Impl::Lib;
    return DnsKey{DnsKey::Flags{flags},
                  DnsKey::Protocol{protocol},
                  DnsKey::Algorithm{algorithm},
                  DnsKey::PublicKey{public_key}};
}

}//namespace {anonymous}

BOOST_AUTO_TEST_CASE(success)
{
    const auto domain = make_absolute_fqdn("iana.org");
    const auto dnskey = make_dnskey(257, 3, 8, iana_org_public_key);
    Fred::Zone::Impl::Lib::WireFormat wire;
    BOOST_CHECK_NO_THROW(wire.append(domain));
    BOOST_CHECK_EQUAL(wire.data().size(), 10);
    BOOST_CHECK_NO_THROW(wire.append(dnskey));
    BOOST_CHECK_EQUAL(wire.data().size(), 10 + 4 + 260);
    const auto hash = Fred::Zone::Impl::Lib::compute_sha256_hash(wire.data().data(), wire.data().size());
    BOOST_CHECK_EQUAL(to_string(hash), "5FB72AE84C35A35DDF7FA24EA791CD9763C4360979C0CBB6FF2C407AF9AE5802");
}

BOOST_AUTO_TEST_SUITE_END()//TestDnsKey
