/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/ds_record.hh"

#include "liblog/liblog.hh"

#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE(TestDsRecord)

// iana.org. DNSKEY 257 3 8 AwEAAb0i1jarxjgmjsOAO7IbXvhwe4yBSGWpUnjj9SLZOguEbOY+6hrs
//                          LL5Xa9LKHDjgF/ovovtOj4t9JrKXQZ6c/kJdtaIkGcWzhACoNkum20mn
//                          2Ja3Zk1xJgEkMpOED5I3N6IHNCALSNw8rxu3c3xVDpD6Tf10+Bcpr4sj
//                          ge32lNeVDjVPs7D7yJSQrr0tBui4rtsMkeQ8jNT9Hna7C/5MRAxqjpue
//                          X0Z7kEqv4vevv1Y7xDi7yloRzY/tx8BdYEVRNOQ9pbU2ERuFhUWmwT1D
//                          ROUBwpK6husdOGbnPyGzCiHt14An5wYpj6B4Jao9DuNTgTl6xg20oyR4
//                          6pRM1roTeIk=
// iana.org. DS 39817 8 2 5FB72AE84C35A35DDF7FA24EA791CD9763C4360979C0CBB6FF2C407A
//                        F9AE5802
// iana.org. DS 39817 8 1 4B2634F69EDD8D0741DA8629E2064B05E8138E51

// nic.cz. DNSKEY 257 3 13 LM4zvjUgZi2XZKsYooDE0HFYGfWp242fKB+O8sLsuox8S6MJTowY8lBD
//                         jZD7JKbmaNot3+1H8zU9TrDzWmmHwQ==
// nic.cz. DS 61281 13 2 4104D40C8FE2030BF7A09A199FCF37B36F7EC8DDD16F5A84F2E61C24
//                       8D3AFD0F
namespace {

const auto iana_org_public_key = Fred::Zone::Impl::Lib::Base64EncodedData{
        "AwEAAb0i1jarxjgmjsOAO7IbXvhwe4yBSGWpUnjj9SLZOguEbOY+6hrs"
        "LL5Xa9LKHDjgF/ovovtOj4t9JrKXQZ6c/kJdtaIkGcWzhACoNkum20mn"
        "2Ja3Zk1xJgEkMpOED5I3N6IHNCALSNw8rxu3c3xVDpD6Tf10+Bcpr4sj"
        "ge32lNeVDjVPs7D7yJSQrr0tBui4rtsMkeQ8jNT9Hna7C/5MRAxqjpue"
        "X0Z7kEqv4vevv1Y7xDi7yloRzY/tx8BdYEVRNOQ9pbU2ERuFhUWmwT1D"
        "ROUBwpK6husdOGbnPyGzCiHt14An5wYpj6B4Jao9DuNTgTl6xg20oyR4"
        "6pRM1roTeIk="};

const auto nic_cz_public_key = Fred::Zone::Impl::Lib::Base64EncodedData{
        "LM4zvjUgZi2XZKsYooDE0HFYGfWp242fKB+O8sLsuox8S6MJTowY8lBD"
        "jZD7JKbmaNot3+1H8zU9TrDzWmmHwQ=="};

auto make_absolute_fqdn(const std::string& fqdn)
{
    return Fred::Zone::Impl::Lib::AbsoluteFqdn{Fred::Zone::Impl::Lib::Fqdn{fqdn}};
}

auto make_dnskey(std::uint16_t flags, std::uint8_t protocol, std::uint8_t algorithm, const Fred::Zone::Impl::Lib::Base64EncodedData& public_key)
{
    using namespace Fred::Zone::Impl::Lib;
    return DnsKey{DnsKey::Flags{flags},
                  DnsKey::Protocol{protocol},
                  DnsKey::Algorithm{algorithm},
                  DnsKey::PublicKey{public_key}};
}

bool log_exception(const std::exception& e)
{
    LIBLOG_INFO("std::exception caught: {}", e.what());
    return true;
}

}//namespace {anonymous}

BOOST_AUTO_TEST_CASE(iana_org)
{
    const auto domain = make_absolute_fqdn("iana.org");
    const auto dnskey = make_dnskey(257, 3, 8, iana_org_public_key);
    const auto key_tag = Fred::Zone::Impl::Lib::DsRecord::make_key_tag(dnskey);
    BOOST_CHECK_EQUAL(*key_tag, 39817);
    auto digest = Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha1"});
    BOOST_CHECK_EQUAL(*digest, "4B2634F69EDD8D0741DA8629E2064B05E8138E51");
    digest = Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha256"});
    BOOST_CHECK_EQUAL(*digest, "5FB72AE84C35A35DDF7FA24EA791CD9763C4360979C0CBB6FF2C407AF9AE5802");
    digest = Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha384"});
    BOOST_CHECK_EQUAL(*digest, "8D4B950BB71C3CEF58A48A925440C3E26BBC213114B646C1ABB058E6A0340B8576FD4AB5AD743C4EC209BCCAB2029239");
    BOOST_CHECK_EXCEPTION((Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha2"})), std::exception, log_exception);
    BOOST_CHECK_EXCEPTION((Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha255"})), std::exception, log_exception);
    BOOST_CHECK_EXCEPTION((Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha385"})), std::exception, log_exception);
}

BOOST_AUTO_TEST_CASE(nic_cz)
{
    const auto domain = make_absolute_fqdn("nic.cz");
    const auto dnskey = make_dnskey(257, 3, 13, nic_cz_public_key);
    const auto key_tag = Fred::Zone::Impl::Lib::DsRecord::make_key_tag(dnskey);
    BOOST_CHECK_EQUAL(*key_tag, 61281);
    auto digest = Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha1"});
    BOOST_CHECK_EQUAL(*digest, "091CECC4D2AADB7AC8C4DF413DDF9C5B0B61E5B6");
    digest = Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha256"});
    BOOST_CHECK_EQUAL(*digest, "4104D40C8FE2030BF7A09A199FCF37B36F7EC8DDD16F5A84F2E61C248D3AFD0F");
    digest = Fred::Zone::Impl::Lib::DsRecord::make_digest(domain, dnskey, Fred::Zone::Impl::Lib::DsRecord::DigestType{"sha384"});
    BOOST_CHECK_EQUAL(*digest, "96AB1359B7542A9CE14728B9307CB8111A7DECEC087202479251D7F318DB29422EBAACC9B526AD6ED0F278D15BE51246");
}

BOOST_AUTO_TEST_SUITE_END()//TestDsRecord
