/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "setup.hh"

#include "src/impl/lib/get_zone_data_context.hh"
#include "src/impl/lib/get_zones.hh"

#include "liblog/liblog.hh"

#include <boost/test/unit_test.hpp>

#include <algorithm>
#include <stdexcept>

BOOST_AUTO_TEST_SUITE(TestZoneData)

namespace {

#if 0
CREATE FUNCTION get_object_state_id(state TEXT)
RETURNS integer AS
'SELECT id FROM enum_object_states WHERE name=state'
LANGUAGE SQL IMMUTABLE RETURNS NULL ON NULL INPUT PARALLEL SAFE
SET search_path='public';

SELECT obr.name, d.keyset, d.nsset
  FROM domain d
  JOIN zone z ON z.id = d.zone AND
                 z.fqdn = 'cz'
  JOIN object_registry obr ON obr.id = d.id
 WHERE d.nsset IS NOT NULL AND
       NOT EXISTS(SELECT FROM object_state os
                         JOIN enum_object_states eos ON eos.id = os.state_id AND
                                                        eos.name = 'outzone'
                        WHERE os.object_id = d.id AND
                              os.valid_to IS NULL);

SELECT DISTINCT dnskey.keysetid, dnskey.flags, dnskey.protocol, dnskey.alg, dnskey.key
  FROM domain d
  JOIN zone z ON z.id = d.zone AND
                 z.fqdn = 'cz'
  JOIN dnskey ON dnskey.keysetid = d.keyset
 WHERE NOT EXISTS(SELECT FROM object_state os
                         JOIN enum_object_states eos ON eos.id = os.state_id AND
                                                        eos.name = 'outzone'
                        WHERE os.object_id = d.id AND
                              os.valid_to IS NULL);

SELECT DISTINCT h.nssetid, h.fqdn, him.ipaddr
  FROM domain d
  JOIN zone z ON z.id = d.zone AND
                 z.fqdn = 'cz'
  JOIN host h ON h.nssetid = d.nsset
  LEFT JOIN host_ipaddr_map him ON him.hostid = h.id AND
                                   him.nssetid = h.nssetid
 WHERE NOT EXISTS(SELECT FROM object_state os
                         JOIN enum_object_states eos ON eos.id = os.state_id AND
                                                        eos.name = 'outzone'
                        WHERE os.object_id = d.id AND
                              os.valid_to IS NULL);
#endif

bool is_subdomain(std::string subdomain, std::string domain)
{
    static const auto make_absolute_fqdn = [](std::string& fqdn)
    {
        if (fqdn.empty() ||
            (fqdn.back() != '.'))
        {
            fqdn.append(1, '.');
        }
    };
    make_absolute_fqdn(subdomain);
    make_absolute_fqdn(domain);
    if ((domain.length() + 1) < subdomain.length())
    {
        const auto subdomain_iter = subdomain.end() - domain.length();
        return (subdomain_iter[-1] == '.') && std::equal(std::begin(domain), std::end(domain), subdomain_iter, std::end(subdomain));
    }
    return false;
}

void test_get_zones()
{
    auto tx = Test::make_transaction();
    const auto zones = Fred::Zone::Impl::Lib::get_zones(tx);
    BOOST_CHECK_LT(0, zones.size());
    std::for_each(begin(zones), end(zones), [](const auto& zone) { BOOST_TEST_MESSAGE(zone); });
    const bool has_cz_zone = find_if(begin(zones), end(zones), [](const auto& zone) { return zone == Fred::Zone::Impl::Lib::ZoneName{"cz"}; }) != end(zones);
    BOOST_REQUIRE(has_cz_zone);
    commit(std::move(tx));
}

void test_zone_data()
{
    auto ctx = Fred::Zone::Impl::Lib::get_zone_data_context({Test::make_connection(),
                                                             Test::make_connection(),
                                                             Test::make_connection()},
                                                            Fred::Zone::Impl::Lib::ZoneName{"cz"});
    Fred::Zone::Impl::Lib::ZoneData::Options options;
    options.ds_record.digest_types = {Fred::Zone::Impl::Lib::ZoneData::DsRData::DigestType{"sha256"}};
    const auto zone_data = Fred::Zone::Impl::Lib::ZoneData{std::move(ctx), options};
    LIBLOG_INFO("Zone data generation start");
    auto lines = std::int64_t{1};//SOA record
    auto glues = std::int64_t{0};
    auto empty_glues = std::int64_t{0};
    auto glue_a_records = std::int64_t{0};
    auto glue_aaaa_records = std::int64_t{0};
    std::for_each(zone_data.begin(), zone_data.end(), [&](const Fred::Zone::Impl::Lib::ZoneData::Domain& domain)
    {
        lines += domain.get_ds_r_data().size() + domain.get_nameservers().size();
        std::for_each(
                std::begin(domain.get_nameservers()),
                std::end(domain.get_nameservers()),
                [&](const Fred::Zone::Impl::Lib::ZoneData::Nameserver& ns)
        {
            if (ns.are_glue_records_permitted())
            {
                ++glues;
                if (!ns.get_glue_records().a_records.empty() ||
                    !ns.get_glue_records().aaaa_records.empty())
                {
                    glue_a_records += ns.get_glue_records().a_records.size();
                    glue_aaaa_records += ns.get_glue_records().aaaa_records.size();
                }
                else
                {
                    ++empty_glues;
                    if (empty_glues <= 100)
                    {
                        BOOST_TEST_MESSAGE("domain: " << domain.get_domain_name().get_fqdn() << ", ns: " << ns.get_nameserver_hostname().get_fqdn()
                                                      << " does not have any A/AAAA records");
                    }
                }
            }
        });
    });
    BOOST_TEST_MESSAGE("zone file has " << lines << " lines");
    BOOST_TEST_MESSAGE("zone file has " << glues << " glues");
    BOOST_TEST_MESSAGE("zone file has " << empty_glues << " empty glues");
    BOOST_TEST_MESSAGE("zone file has " << glue_a_records << " glue A records");
    BOOST_TEST_MESSAGE("zone file has " << glue_aaaa_records << " glue AAAA records");
    LIBLOG_INFO("Zone data generation done");
}

}//namespace {anonymous}

BOOST_AUTO_TEST_CASE(test_glue_records)
{
    BOOST_CHECK(is_subdomain("ns.seznam.cz", "seznam.cz"));
    BOOST_CHECK(is_subdomain("ns.seznam.cz", "seznam.cz."));
    BOOST_CHECK(is_subdomain("ns.seznam.cz.", "seznam.cz"));
    BOOST_CHECK(is_subdomain("ns.seznam.cz.", "seznam.cz."));
    BOOST_CHECK(!is_subdomain("seznam.cz", "seznam.cz"));
    BOOST_CHECK(!is_subdomain("seznam.cz", "ns.seznam.cz"));
    BOOST_CHECK(!is_subdomain(".seznam.cz", "seznam.cz"));
    BOOST_CHECK(!is_subdomain("ns.seznam.cz", "seynam.cz"));
    BOOST_CHECK(!is_subdomain("ns.seznam.cz", "Seznam.cz"));
}

BOOST_AUTO_TEST_CASE(test_all)
{
    BOOST_CHECK_NO_THROW(test_get_zones());
    BOOST_CHECK_NO_THROW(test_zone_data());
}

BOOST_AUTO_TEST_SUITE_END()//TestZoneData
